# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cgi
from datetime import datetime
import imghdr
from io import StringIO
import os
import tempfile
from unittest import mock
import uuid

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from webob import exc

from murano.api.v1 import catalog
from murano.api.v1 import PKG_PARAMS_MAP
from murano.common import exceptions as common_exc
from murano.db.catalog import api as db_catalog_api
from murano.db import models
from murano.packages import exceptions
from murano.packages import load_utils
import murano.tests.unit.api.base as test_base
import murano.tests.unit.utils as test_utils

CONF = cfg.CONF


class TestCatalogApi(test_base.ControllerTest, test_base.MuranoApiTestCase):

    def setUp(self):
        super(TestCatalogApi, self).setUp()
        self.controller = catalog.Controller()
        _, self.test_package = self._test_package()

    def _add_pkg(self, tenant_name, public=False, classes=None, **kwargs):
        package_to_upload = self.test_package.copy()
        package_to_upload['is_public'] = public
        package_to_upload['fully_qualified_name'] = str(uuid.uuid4())
        if classes:
            package_to_upload['class_definitions'] = classes
        else:
            package_to_upload['class_definitions'] = []
        package_to_upload.update(kwargs)
        return db_catalog_api.package_upload(
            package_to_upload, tenant_name)

    def test_packages_filtering_admin(self):
        self.is_admin = True
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        for dummy in range(7):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')

        pkg = self._add_pkg('test_tenant', type='Library')
        self._add_pkg('test_tenant')
        self._add_pkg('test_tenant2', public=True, type='Library')
        self._add_pkg('test_tenant3')

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'False',
                                             'owned': 'False'}))
        self.assertEqual(4, len(result['packages']))
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'False',
                                             'owned': 'True'}))
        self.assertEqual(2, len(result['packages']))

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True',
                                             'owned': 'False'}))
        self.assertEqual(3, len(result['packages']))
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True',
                                             'owned': 'True'}))
        self.assertEqual(2, len(result['packages']))

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'owned': 'True',
                'fqn': pkg.fully_qualified_name}))
        self.assertEqual(1, len(result['packages']))
        self.assertEqual(pkg.fully_qualified_name,
                         result['packages'][0]['fully_qualified_name'])

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'owned': 'True',
                'type': 'Library'}))
        self.assertEqual(1, len(result['packages']))
        self.assertEqual(pkg.fully_qualified_name,
                         result['packages'][0]['fully_qualified_name'])

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'type': 'Library'}))
        self.assertEqual(2, len(result['packages']))

    def test_packages_filtering_non_admin(self):
        self.is_admin = False
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        for dummy in range(8):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')

        pkg = self._add_pkg('test_tenant', type='Library')
        self._add_pkg('test_tenant')
        self._add_pkg('test_tenant2', public=True, type='Library')
        self._add_pkg('test_tenant3')

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'False',
                                             'owned': 'False'}))
        self.assertEqual(3, len(result['packages']))
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'False',
                                             'owned': 'True'}))
        self.assertEqual(2, len(result['packages']))
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True',
                                             'owned': 'False'}))
        self.assertEqual(3, len(result['packages']))
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True',
                                             'owned': 'True'}))
        self.assertEqual(2, len(result['packages']))

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'owned': 'True',
                'fqn': pkg.fully_qualified_name}))
        self.assertEqual(1, len(result['packages']))
        self.assertEqual(pkg.fully_qualified_name,
                         result['packages'][0]['fully_qualified_name'])

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'owned': 'True',
                'type': 'Library'}))
        self.assertEqual(1, len(result['packages']))
        self.assertEqual(pkg.fully_qualified_name,
                         result['packages'][0]['fully_qualified_name'])

        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={
                'type': 'Library'}))
        self.assertEqual(2, len(result['packages']))

        self._set_policy_rules({'get_package': '',
                                'manage_public_package': '!'})
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'False'}))
        self.assertEqual(2, len(result['packages']))

    def test_packages_filter_by_id(self):
        """Test that packages are filtered by ID

        GET /catalog/packages with parameter "id" returns packages
        filtered by id.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        _, package1_data = self._test_package()
        _, package2_data = self._test_package()

        package1_data['fully_qualified_name'] += '_1'
        package1_data['name'] += '_1'
        package1_data['class_definitions'] = (u'test.mpl.v1.app.Thing1',)
        package2_data['fully_qualified_name'] += '_2'
        package2_data['name'] += '_2'
        package2_data['class_definitions'] = (u'test.mpl.v1.app.Thing2',)

        expected_package = db_catalog_api.package_upload(package1_data, '')
        db_catalog_api.package_upload(package2_data, '')

        req = self._get('/catalog/packages',
                        params={'id': expected_package.id})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        res = req.get_response(self.api)
        self.assertEqual(200, res.status_code)

        self.assertEqual(1, len(res.json['packages']))

        found_package = res.json['packages'][0]

        self.assertEqual(expected_package.id, found_package['id'])

    def test_packages_filter_by_in_category(self):
        """Test that packages are filtered by in:cat1,cat2,cat3

        GET /catalog/packages with parameter "category=in:cat1,cat2,cat3"
        returns packages filtered by category.
        """
        names = ['cat1', 'cat2', 'cat3', 'cat4']
        for name in names:
            db_catalog_api.category_add(name)
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        _, package1_data = self._test_package()
        _, package2_data = self._test_package()
        _, package3_data = self._test_package()

        package1_data['fully_qualified_name'] += '_1'
        package1_data['name'] += '_1'
        package1_data['class_definitions'] = (u'test.mpl.v1.app.Thing1',)
        package1_data['categories'] = ('cat1', 'cat2')

        package2_data['fully_qualified_name'] += '_2'
        package2_data['name'] += '_2'
        package2_data['class_definitions'] = (u'test.mpl.v1.app.Thing2',)
        package2_data['categories'] = ('cat2', 'cat3')

        package3_data['fully_qualified_name'] += '_3'
        package3_data['name'] += '_3'
        package3_data['class_definitions'] = (u'test.mpl.v1.app.Thing3',)
        package3_data['categories'] = ('cat2', 'cat4')

        expected_packages = [db_catalog_api.package_upload(package1_data, ''),
                             db_catalog_api.package_upload(package2_data, '')]
        db_catalog_api.package_upload(package3_data, '')

        categories_in = "in:cat1,cat3"

        req = self._get('/catalog/packages',
                        params={'category': categories_in})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        res = req.get_response(self.api)
        self.assertEqual(200, res.status_code)
        self.assertEqual(2, len(res.json['packages']))

        found_packages = res.json['packages']

        self.assertEqual([pack.id for pack in expected_packages],
                         [pack['id'] for pack in found_packages])

    def test_packages_filter_by_in_tag(self):
        """Test that packages are filtered by in:tag1,tag2,tag3

        GET /catalog/packages with parameter "tag=in:tag1,tag2,tag3"
        returns packages filtered by category.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        _, package1_data = self._test_package()
        _, package2_data = self._test_package()
        _, package3_data = self._test_package()

        package1_data['fully_qualified_name'] += '_1'
        package1_data['name'] += '_1'
        package1_data['class_definitions'] = (u'test.mpl.v1.app.Thing1',)
        package1_data['tags'] = ('tag1', 'tag2')

        package2_data['fully_qualified_name'] += '_2'
        package2_data['name'] += '_2'
        package2_data['class_definitions'] = (u'test.mpl.v1.app.Thing2',)
        package2_data['tags'] = ('tag2', 'tag3')

        package3_data['fully_qualified_name'] += '_3'
        package3_data['name'] += '_3'
        package3_data['class_definitions'] = (u'test.mpl.v1.app.Thing3',)
        package3_data['tags'] = ('tag2', 'tag4')

        expected_packages = [db_catalog_api.package_upload(package1_data, ''),
                             db_catalog_api.package_upload(package2_data, '')]
        db_catalog_api.package_upload(package3_data, '')

        tag_in = "in:tag1,tag3"

        req = self._get('/catalog/packages',
                        params={'tag': tag_in})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        res = req.get_response(self.api)
        self.assertEqual(200, res.status_code)
        self.assertEqual(2, len(res.json['packages']))

        found_packages = res.json['packages']

        self.assertEqual([pack.id for pack in expected_packages],
                         [pack['id'] for pack in found_packages])

    def test_packages_filter_by_in_id(self):
        """Test that packages are filtered by in:id1,id2,id3

        GET /catalog/packages with parameter "id=in:id1,id2" returns packages
        filtered by id.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        _, package1_data = self._test_package()
        _, package2_data = self._test_package()
        _, package3_data = self._test_package()

        package1_data['fully_qualified_name'] += '_1'
        package1_data['name'] += '_1'
        package1_data['class_definitions'] = (u'test.mpl.v1.app.Thing1',)
        package2_data['fully_qualified_name'] += '_2'
        package2_data['name'] += '_2'
        package2_data['class_definitions'] = (u'test.mpl.v1.app.Thing2',)
        package3_data['fully_qualified_name'] += '_3'
        package3_data['name'] += '_3'
        package3_data['class_definitions'] = (u'test.mpl.v1.app.Thing3',)

        expected_packages = [db_catalog_api.package_upload(package1_data, ''),
                             db_catalog_api.package_upload(package2_data, '')]
        db_catalog_api.package_upload(package3_data, '')

        id_in = "in:" + ",".join(pack.id for pack in expected_packages)

        req = self._get('/catalog/packages',
                        params={'id': id_in})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        res = req.get_response(self.api)
        self.assertEqual(200, res.status_code)

        self.assertEqual(2, len(res.json['packages']))

        found_packages = res.json['packages']

        self.assertEqual([pack.id for pack in expected_packages],
                         [pack['id'] for pack in found_packages])

    def test_packages_filter_by_in_id_empty(self):
        """Test that packages are filtered by "id=in:"

        GET /catalog/packages with parameter "id=in:" returns packages
        filtered by id, in this case no packages should be returned.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        _, package1_data = self._test_package()

        db_catalog_api.package_upload(package1_data, '')

        req = self._get('/catalog/packages', params={'id': "in:"})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        res = req.get_response(self.api)
        self.assertEqual(200, res.status_code)

        self.assertEqual(0, len(res.json['packages']))

    def test_packages_filter_by_name(self):
        """Test that packages are filtered by name

        GET /catalog/packages with parameter "name" returns packages
        filtered by name.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )

        expected_pkg1 = self._add_pkg("test_tenant", name="test_pkgname_1")
        expected_pkg2 = self._add_pkg("test_tenant", name="test_pkgname_2")

        req_pkgname1 = self._get('/catalog/packages',
                                 params={'name': expected_pkg1.name})
        req_pkgname2 = self._get('/catalog/packages',
                                 params={'name': expected_pkg2.name})

        for dummy in range(2):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')

        res_pkgname1 = req_pkgname1.get_response(self.api)
        self.assertEqual(200, res_pkgname1.status_code)
        self.assertEqual(1, len(res_pkgname1.json['packages']))
        self.assertEqual(expected_pkg1.name,
                         res_pkgname1.json['packages'][0]['name'])

        res_pkgname2 = req_pkgname2.get_response(self.api)
        self.assertEqual(200, res_pkgname2.status_code)
        self.assertEqual(1, len(res_pkgname2.json['packages']))
        self.assertEqual(expected_pkg2.name,
                         res_pkgname2.json['packages'][0]['name'])

    def test_packages_filter_by_type(self):
        """Test that packages are filtered by type

        GET /catalog/packages with parameter "type" returns packages
        filtered by type.
        """
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        excepted_pkg1 = self._add_pkg("test_tenant", type='Library')
        excepted_pkg2 = self._add_pkg("test_tenant", type='Library')
        excepted_pkg3 = self._add_pkg("test_tenant", type='Application')

        # filter by type=Library can see 2 pkgs
        req_lib = self._get('/catalog/packages',
                            params={'type': 'Library'})

        # filter by type=Application only see 1 pkgs
        req_app = self._get('/catalog/packages',
                            params={'type': 'Application'})

        for dummy in range(2):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')

        res_lib = req_lib.get_response(self.api)

        self.assertEqual(200, res_lib.status_code)
        self.assertEqual(2, len(res_lib.json['packages']))

        self.assertEqual('Library', res_lib.json['packages'][0]['type'])
        self.assertEqual(excepted_pkg1.name,
                         res_lib.json['packages'][0]['name'])

        self.assertEqual('Library', res_lib.json['packages'][1]['type'])
        self.assertEqual(excepted_pkg2.name,
                         res_lib.json['packages'][1]['name'])

        res_app = req_app.get_response(self.api)

        self.assertEqual(200, res_app.status_code)
        self.assertEqual(1, len(res_app.json['packages']))

        self.assertEqual('Application', res_app.json['packages'][0]['type'])
        self.assertEqual(excepted_pkg3.name,
                         res_app.json['packages'][0]['name'])

    @mock.patch('murano.api.v1.catalog.LOG')
    def test_packages_filter_by_order_by(self, mock_log):
        warnings = []
        mock_log.warning = lambda msg: warnings.append(msg)

        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        self._add_pkg("test_tenant", type='Library')

        # Test whether a valid order by value works.
        order_by = 'name'
        request = self._get('/catalog/packages',
                            params={'order_by': order_by})

        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        self.controller.search(request)
        self.assertEqual(len(warnings), 0)

        # Test whether an invalid order by value fails.
        order_by = 'TEST ORDER BY'
        request = self._get('/catalog/packages', params={'order_by': order_by})

        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        self.controller.search(request)
        self.assertEqual(len(warnings), 1)
        self.assertIn('parameter is not valid', warnings[0])

    def test_packages_filter_by_limit(self):
        """Test that packages are filtered by limit."""
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        pkg = self._add_pkg("test_tenant", type='Library')

        request = self._get('/catalog/packages', params={'limit': '1'})
        self.expect_policy_check('get_package')

        self.expect_policy_check('manage_public_package')
        res = self.controller.search(request)

        self.assertIn('next_marker', res)
        self.assertEqual(res['next_marker'], pkg['id'])

    def test_packages_filter_by_limit_negative_cases(self):
        """Test whether invalid limit values throw expected exceptions."""
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        self._add_pkg("test_tenant", type='Library')

        # Test wheter non-number value throws exception
        request = self._get('/catalog/packages',
                            params={'limit': 'not a number'})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.search,
                              request)
        self.assertEqual(e.explanation, 'Limit param must be an integer')

        # Test whether below-zero value throws exception
        request = self._get('/catalog/packages', params={'limit': '-1'})
        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.search,
                              request)
        self.assertEqual(e.explanation, 'Limit param must be positive')

    @mock.patch('murano.common.utils.split_for_quotes')
    @mock.patch('murano.api.v1.catalog.LOG')
    def test_packages_filter_handle_value_error(self, mock_log, mock_func):
        warnings = []
        mock_func.side_effect = ValueError
        mock_log.warning = lambda msg: warnings.append(msg)

        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        self._add_pkg("test_tenant", type='Library')
        tag_in = "in:tag1,tag3"
        self._get('/catalog/packages', params={'tag': tag_in})
        request = self._get('/catalog/packages',
                            params={'tag': tag_in})

        self.expect_policy_check('get_package')
        self.expect_policy_check('manage_public_package')

        self.controller.search(request)
        self.assertEqual(len(warnings), 1)
        self.assertIn("Search by parameter 'tag' caused an  error",
                      warnings[0])

    def test_packages_filter_by_search(self):
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        excepted_pkg1 = self._add_pkg("test_tenant",
                                      type='Application',
                                      name='awcloud',
                                      description='some context')
        excepted_pkg2 = self._add_pkg("test_tenant",
                                      type='Application',
                                      name='mysql',
                                      description='awcloud product')
        excepted_pkg3 = self._add_pkg("test_tenant",
                                      type='Application',
                                      name='package3',
                                      author='mysql author')

        # filter by search=awcloud can see 2 pkgs
        req_awc = self._get('/catalog/packages',
                            params={'search': 'awcloud'})

        # filter by search=mysql only see 2 pkgs
        req_mysql = self._get('/catalog/packages',
                              params={'search': 'mysql'})

        for dummy in range(2):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')

        res_awc = req_awc.get_response(self.api)

        self.assertEqual(200, res_awc.status_code)
        self.assertEqual(2, len(res_awc.json['packages']))

        self.assertEqual(excepted_pkg1.name,
                         res_awc.json['packages'][0]['name'])

        self.assertEqual(excepted_pkg2.name,
                         res_awc.json['packages'][1]['name'])

        res_mysql = req_mysql.get_response(self.api)

        self.assertEqual(200, res_mysql.status_code)
        self.assertEqual(2, len(res_mysql.json['packages']))

        self.assertEqual(excepted_pkg2.name,
                         res_mysql.json['packages'][0]['name'])
        self.assertEqual(excepted_pkg3.name,
                         res_mysql.json['packages'][1]['name'])

    def test_packages(self):
        self._set_policy_rules(
            {'get_package': '',
             'manage_public_package': ''}
        )
        for dummy in range(9):
            self.expect_policy_check('get_package')
            self.expect_policy_check('manage_public_package')
        result = self.controller.search(self._get('/v1/catalog/packages/'))
        self.assertEqual(0, len(result['packages']))

        self._add_pkg('test_tenant')
        self._add_pkg('test_tenant')
        self._add_pkg('other_tenant')
        self._add_pkg('other_tenant')

        # non-admin should only see 2 pkgs he can edit.
        self.is_admin = False
        result = self.controller.search(self._get('/v1/catalog/packages/'))
        self.assertEqual(2, len(result['packages']))
        # can only deploy his + public
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True'}))
        self.assertEqual(2, len(result['packages']))

        # admin can edit anything
        self.is_admin = True
        result = self.controller.search(self._get('/v1/catalog/packages/'))
        self.assertEqual(4, len(result['packages']))
        # admin can only deploy his + public
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True'}))
        self.assertEqual(2, len(result['packages']))

        self._add_pkg('test_tenant', public=True)
        self._add_pkg('other_tenant', public=True)

        # non-admin are allowed to edit public packages by policy
        self.is_admin = False
        result = self.controller.search(self._get('/v1/catalog/packages/'))
        self.assertEqual(4, len(result['packages']))
        # can deploy mine + other public
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True'}))
        self.assertEqual(4, len(result['packages']))

        # admin can edit anything
        self.is_admin = True
        result = self.controller.search(self._get('/v1/catalog/packages/'))
        self.assertEqual(6, len(result['packages']))
        # can deploy mine + public
        result = self.controller.search(self._get(
            '/v1/catalog/packages/', params={'catalog': 'True'}))
        self.assertEqual(4, len(result['packages']))

    def _test_package(self, manifest='manifest.yaml'):
        package_dir = os.path.abspath(
            os.path.join(
                __file__,
                '../../../packages/test_packages/test.mpl.v1.app',
            )
        )
        pkg = load_utils.load_from_dir(
            package_dir, filename=manifest
        )
        package = {
            'fully_qualified_name': pkg.full_name,
            'type': pkg.package_type,
            'author': pkg.author,
            'supplier': pkg.supplier,
            'name': pkg.display_name,
            'description': pkg.description,
            'is_public': True,
            'tags': pkg.tags,
            'logo': pkg.logo,
            'supplier_logo': pkg.supplier_logo,
            'ui_definition': pkg.ui,
            'class_definitions': tuple(pkg.classes),
            'archive': pkg.blob,
            'categories': [],
        }
        return pkg, package

    def test_modify_package_tags(self):
        self._set_policy_rules(
            {'modify_package': '',
             'manage_public_package': '',
             'publicize_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        saved_package_pub = self._add_pkg('test_tenant', public=True)
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})

        url = '/v1/catalog/packages/' + str(saved_package.id)

        data = []
        data.append({'op': 'add', 'path': ['tags'], 'value': ["foo", "bar"]})

        req = self._patch(url, jsonutils.dump_as_bytes(data))
        result = self.controller.update(req, data, saved_package.id)

        self.assertIn('foo', result['tags'])
        self.assertIn('bar', result['tags'])

        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package_pub.id})
        self.expect_policy_check('manage_public_package', {})
        url = '/v1/catalog/packages/' + str(saved_package_pub.id)

        data = []
        data.append({'op': 'add', 'path': ['tags'], 'value': ["foo", "bar"]})

        req = self._patch(url, jsonutils.dump_as_bytes(data))
        result = self.controller.update(req, data, saved_package_pub.id)

        self.assertIn('foo', result['tags'])
        self.assertIn('bar', result['tags'])

    def test_modify_package_name(self):
        self._set_policy_rules(
            {'modify_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})
        url = '/v1/catalog/packages/' + str(saved_package.id)

        data = []
        data.append({'op': 'replace', 'path': ['name'], 'value': 'test_name'})
        req = self._patch(url, jsonutils.dump_as_bytes(data))
        result = self.controller.update(req, data, saved_package.id)
        self.assertEqual('test_name', result['name'])
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})

        data = []
        data.append({'op': 'replace', 'path': ['name'], 'value': 'a' * 81})
        req = self._patch(url, jsonutils.dump_as_bytes(data))
        self.assertRaises(exc.HTTPBadRequest, self.controller.update,
                          req, data, saved_package.id)

    def test_modify_package_no_body(self):
        self._set_policy_rules(
            {'modify_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})
        url = '/v1/catalog/packages/' + str(saved_package.id)
        req = self._patch(url, jsonutils.dump_as_bytes(None))
        self.assertRaises(exc.HTTPBadRequest, self.controller.update,
                          req, None, saved_package.id)

    def test_modify_package_is_public(self):
        self._set_policy_rules(
            {'modify_package': '',
             'manage_public_package': '',
             'publicize_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        url = '/v1/catalog/packages/' + str(saved_package.id)
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})
        self.expect_policy_check('publicize_package', {})
        data = []
        data.append({'op': 'replace', 'path': ['is_public'], 'value': True})
        req = self._patch(url, jsonutils.dump_as_bytes(data))
        result = self.controller.update(req, data, saved_package.id)
        self.assertTrue(result['is_public'])

    def test_modify_package_bad_content_type(self):
        self._set_policy_rules(
            {'modify_package': '',
             'manage_public_package': '',
             'publicize_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        url = '/v1/catalog/packages/' + str(saved_package.id)
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})
        self.expect_policy_check('publicize_package', {})
        data = []
        data.append({'op': 'replace', 'path': ['is_public'], 'value': True})
        req = self._patch(url, jsonutils.dump_as_bytes(data))
        result = self.controller.update(req, data, saved_package.id)
        self.assertTrue(result['is_public'])

    def test_modify_package_with_incorrect_content_type(self):
        self._set_policy_rules(
            {'modify_package': ''}
        )
        saved_package = self._add_pkg('test_tenant')
        self.expect_policy_check('modify_package',
                                 {'package_id': saved_package.id})
        url = '/v1/catalog/packages/' + str(saved_package.id)
        data = []
        data.append({'op': 'replace', 'path': ['name'], 'value': 'test_name'})

        req = self._patch(url, jsonutils.dump_as_bytes(data))
        req.get_content_type = mock.MagicMock(
            side_effect=common_exc.UnsupportedContentType)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.update, req, data, saved_package.id)

    def test_not_valid_logo(self):
        self.assertRaises(exceptions.PackageLoadError,
                          self._test_package, 'manifest_with_broken_logo.yaml')

    def test_load_package_with_supplier_info(self):
        self._set_policy_rules(
            {'get_package': '@'}
        )
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        self.expect_policy_check('get_package',
                                 {'package_id': saved_package.id})

        req = self._get('/v1/catalog/packages/%s' % saved_package.id)
        result = self.controller.get(req, saved_package.id)

        self.assertEqual(package['supplier'], result['supplier'])

        req = self._get(
            '/v1/catalog/packages/%s/supplier_logo' % saved_package.id
        )
        result = self.controller.get_supplier_logo(req, saved_package.id)

        self.assertEqual('png', imghdr.what('', result))

    def test_download_package(self):
        self._set_policy_rules(
            {'download_package': '@'}
        )
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        self.expect_policy_check('download_package',
                                 {'package_id': saved_package.id})

        req = self._get_with_accept('/catalog/packages/%s/download'
                                    % saved_package.id,
                                    accept='application/octet-stream')

        result = req.get_response(self.api)

        self.assertEqual(200, result.status_code)

    def test_download_package_negative(self):

        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        req = self._get_with_accept('/catalog/packages/%s/download'
                                    % saved_package.id,
                                    accept='application/foo')

        result = req.get_response(self.api)

        self.assertEqual(406, result.status_code)
        self.assertIn(b'Acceptable response can not be provided',
                      result.body)

    @mock.patch('murano.api.v1.catalog._validate_body')
    @mock.patch('murano.common.policy.check')
    def test_upload_package_with_invalid_schema(self, mock_policy_check,
                                                mock_validate_body):
        invalid_pkg_upload_schema = {"type": None}
        mock_policy_check.return_value = True
        mock_validate_body.return_value = (None, invalid_pkg_upload_schema)
        mock_request = mock.MagicMock(context={})
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                              mock_request)
        self.assertIn(
            "Package schema is not valid",
            e.explanation
        )

    @mock.patch('murano.api.v1.catalog._validate_body')
    @mock.patch('murano.common.policy.check')
    def test_upload_package_with_invalid_file_content(self, mock_policy_check,
                                                      mock_validate_body):
        with tempfile.NamedTemporaryFile(delete=True) as temp_file:
            mock_policy_check.return_value = True
            mock_validate_body.return_value = (mock.MagicMock(file=temp_file),
                                               None)
            mock_request = mock.MagicMock(context={})
            e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                                  mock_request)
            self.assertIn(
                "Uploading file can't be empty",
                e.explanation
            )

    @mock.patch('murano.api.v1.catalog.PKG_PARAMS_MAP')
    @mock.patch('murano.packages.load_utils.load_from_file')
    @mock.patch('murano.api.v1.catalog._validate_body')
    @mock.patch('murano.common.policy.check')
    def test_upload_package_with_oversized_file_name(self, mock_policy_check,
                                                     mock_validate_body,
                                                     mock_load_from_file,
                                                     mock_pkg_params_map):
        mock_policy_check.return_value = True
        mock_load_from_file.return_value = mock.MagicMock(
            __exit__=lambda obj, type, value, tb: False)
        mock_pkg_params_map.return_value = {}
        mock_request = mock.MagicMock(context=mock.MagicMock(
            tenant=self.tenant))
        test_package_meta = {'name': 'a' * 81}
        with tempfile.NamedTemporaryFile(delete=True) as temp_file:
            temp_file.write(b"Random test content\n")
            temp_file.seek(0)
            mock_validate_body.return_value = \
                (mock.MagicMock(file=temp_file), test_package_meta)
            e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                                  mock_request)
            self.assertIn(
                "Package name should be 80 characters maximum",
                e.explanation
            )

    @mock.patch('murano.packages.load_utils.load_from_file')
    @mock.patch('murano.api.v1.catalog._validate_body')
    @mock.patch('murano.common.policy.check')
    def test_upload_package_handle_package_load_error(self, mock_policy_check,
                                                      mock_validate_body,
                                                      mock_load_from_file):
        pkg_to_upload = mock.MagicMock(
            __exit__=lambda obj, type, value, tb: False)
        mock_request = mock.MagicMock(context=mock.MagicMock(
            tenant=self.tenant))
        mock_load_from_file.return_value = pkg_to_upload
        mock_load_from_file.side_effect = exceptions.PackageLoadError
        mock_policy_check.return_value = True
        with tempfile.NamedTemporaryFile(delete=True) as temp_file:
            temp_file.write(b"Random test content\n")
            temp_file.seek(0)
            mock_validate_body.return_value = \
                (mock.MagicMock(file=temp_file), None)
            e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                                  mock_request)
            self.assertIn(
                "Couldn't load package from file",
                e.explanation
            )

    @mock.patch('murano.packages.load_utils.load_from_file')
    @mock.patch('murano.api.v1.catalog._validate_body')
    @mock.patch('murano.common.policy.check')
    def test_upload_package_handle_duplicate_exception(self, mock_policy_check,
                                                       mock_validate_body,
                                                       mock_load_from_file):
        """Test whether duplicate error is correctly thrown."""
        # Save the first package entry to the DB
        test_package_meta = self.test_package.copy()
        test_package_meta['name'] = 'test_package'
        test_package_meta['fully_qualified_name'] = str(uuid.uuid4())
        test_package_meta['description'] = 'test_description'
        test_package_meta['enabled'] = False
        test_package_meta['is_public'] = False
        db_catalog_api.package_upload(test_package_meta, self.tenant)

        # Reverse the operation performed by upload for copying values from
        # pkg_to_upload into package_meta dict.
        pkg_to_upload = mock.MagicMock(
            __exit__=lambda obj, type, value, tb: False)
        for k, v in PKG_PARAMS_MAP.items():
            if v in test_package_meta.keys():
                val = test_package_meta[v]
                setattr(pkg_to_upload.__enter__(), k, val)

        # Delete extra properties so validation in upload passes.
        for attr in ['fully_qualified_name', 'ui_definition', 'author',
                     'supplier_logo', 'supplier', 'logo', 'type', 'archive']:
            if attr in test_package_meta.keys():
                del test_package_meta[attr]

        mock_request = mock.MagicMock(context=mock.MagicMock(
            tenant=self.tenant))
        mock_load_from_file.return_value = pkg_to_upload
        mock_policy_check.return_value = True

        with tempfile.NamedTemporaryFile(delete=True) as temp_file:
            temp_file.write(b"Random test content\n")
            temp_file.seek(0)
            mock_validate_body.return_value = \
                (mock.MagicMock(file=temp_file), test_package_meta)
            e = self.assertRaises(exc.HTTPConflict, self.controller.upload,
                                  mock_request)
            self.assertIn(
                "Package with specified full name is already registered",
                e.detail
            )

    @mock.patch('murano.common.policy.check')
    def test_upload_package_with_oversized_body(self, mock_policy_check):
        mock_policy_check.return_value = True
        packages_to_upload = {'a': 0, 'b': 1, 'c': 2}
        mock_request = mock.MagicMock(context={})
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                              mock_request, body=packages_to_upload)
        self.assertIn(
            "'multipart/form-data' request body should contain 1 or 2 "
            "parts: json string and zip archive.",
            e.explanation
        )

    @mock.patch('murano.common.policy.check')
    def test_upload_package_with_empty_body(self, mock_policy_check):
        mock_policy_check.return_value = True
        packages_to_upload = {}
        mock_request = mock.MagicMock(context={})
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.upload,
                              mock_request, body=packages_to_upload)
        self.assertEqual(
            'There is no file package with application description',
            e.explanation)

    def test_get_ui_definition(self):
        self._set_policy_rules(
            {'get_package': '@'}
        )
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        self.expect_policy_check('get_package',
                                 {'package_id': saved_package.id})

        req = self._get_with_accept('/catalog/packages/%s/ui'
                                    % saved_package.id,
                                    accept="text/plain")

        result = req.get_response(self.api)

        self.assertEqual(200, result.status_code)

    def test_get_ui_definition_negative(self):
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        req = self._get_with_accept('/catalog/packages/%s/ui'
                                    % saved_package.id,
                                    accept='application/foo')

        result = req.get_response(self.api)

        self.assertEqual(406, result.status_code)
        self.assertIn(b'Acceptable response can not be provided',
                      result.body)

    def test_get_logo(self):
        self._set_policy_rules(
            {'get_package': '@'}
        )
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        self.expect_policy_check('get_package',
                                 {'package_id': saved_package.id})

        req = self._get_with_accept('/catalog/packages/%s/logo'
                                    % saved_package.id,
                                    accept="application/octet-stream")

        result = req.get_response(self.api)

        self.assertEqual(200, result.status_code)
        self.assertEqual(package['logo'], result.body)

    def test_get_logo_negative(self):
        _, package = self._test_package()

        saved_package = db_catalog_api.package_upload(package, '')

        req = self._get_with_accept('/catalog/packages/%s/logo'
                                    % saved_package.id,
                                    accept='application/foo')

        result = req.get_response(self.api)

        self.assertEqual(406, result.status_code)
        self.assertIn(b'Acceptable response can not be provided',
                      result.body)

    def test_add_public_unauthorized(self):
        self._set_policy_rules({
            'upload_package': '@',
            'publicize_package': 'is_admin:True',
            'delete_package': 'is_admin:True',
        })

        self.expect_policy_check('upload_package')
        self.expect_policy_check('delete_package', mock.ANY)
        self.expect_policy_check('upload_package')
        self.expect_policy_check('publicize_package')
        self.expect_policy_check('upload_package')
        self.expect_policy_check('publicize_package')

        file_obj_str = StringIO("This is some dummy data")
        file_obj = mock.MagicMock(cgi.FieldStorage)
        file_obj.file = file_obj_str
        package_from_dir, _ = self._test_package()

        body_fmt = '''\

--BOUNDARY
Content-Disposition: form-data; name="__metadata__"

{0}
--BOUNDARY
Content-Disposition: form-data; name="ziparchive"; filename="file.zip"

This is a fake zip archive
--BOUNDARY--'''

        def format_body(content):
            content = jsonutils.dumps(content)
            body = body_fmt.format(content)
            body = body.encode('utf-8')
            return body

        with mock.patch('murano.packages.load_utils.load_from_file') as lff:
            ctxmgr = mock.Mock()
            ctxmgr.__enter__ = mock.Mock(return_value=package_from_dir)
            ctxmgr.__exit__ = mock.Mock(return_value=False)
            lff.return_value = ctxmgr

            # Uploading a non-public package
            req = self._post(
                '/catalog/packages',
                format_body({'is_public': False}),
                content_type='multipart/form-data; ; boundary=BOUNDARY',
            )
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_code)

            self.is_admin = True
            app_id = jsonutils.loads(res.body)['id']
            req = self._delete('/catalog/packages/{0}'.format(app_id))
            res = req.get_response(self.api)

            self.is_admin = False
            # Uploading a public package fails
            req = self._post(
                '/catalog/packages',
                format_body({'is_public': True}),
                content_type='multipart/form-data; ; boundary=BOUNDARY',
            )
            res = req.get_response(self.api)
            self.assertEqual(403, res.status_code)

            # Uploading a public package passes for admin
            self.is_admin = True
            req = self._post(
                '/catalog/packages',
                format_body({'is_public': True}),
                content_type='multipart/form-data; ; boundary=BOUNDARY',
            )
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_code)

    def test_upload_pkg_with_tag(self):
        """Check upload package with tags successfully"""

        self._set_policy_rules({'upload_package': '@'})
        self.expect_policy_check('upload_package')
        file_obj_str = StringIO("This is some dummy data")
        file_obj = mock.MagicMock(cgi.FieldStorage)
        file_obj.file = file_obj_str
        package_from_dir, _ = self._test_package()

        body_fmt = '''\

--BOUNDARY
Content-Disposition: form-data; name="__metadata__"

{0}
--BOUNDARY
Content-Disposition: form-data; name="ziparchive"; filename="file.zip"

This is a fake zip archive
--BOUNDARY--'''

        def format_body(content):
            content = jsonutils.dumps(content)
            body = body_fmt.format(content)
            body = body.encode('utf-8')
            return body

        with mock.patch('murano.packages.load_utils.load_from_file') as lff:
            ctxmgr = mock.Mock()
            ctxmgr.__enter__ = mock.Mock(return_value=package_from_dir)
            ctxmgr.__exit__ = mock.Mock(return_value=False)
            lff.return_value = ctxmgr

            # Uploading a package with foo_tag
            req = self._post(
                '/catalog/packages',
                format_body({'tags': ['foo_tag']}),
                content_type='multipart/form-data; ; boundary=BOUNDARY',
            )
            res = req.get_response(self.api)

            processed_result = jsonutils.loads(res.body)
            # check user set foo_tag in result
            self.assertIn('foo_tag', processed_result["tags"])
            # check tag Linux from package in result
            self.assertIn('Linux', processed_result["tags"])

    def test_add_category(self):
        """Check that category added successfully"""

        self._set_policy_rules({'add_category': '@'})
        self.expect_policy_check('add_category')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        expected = {
            'name': 'new_category',
            'created': datetime.isoformat(fake_now)[:-7],
            'updated': datetime.isoformat(fake_now)[:-7],
            'package_count': 0,
        }

        body = {'name': 'new_category'}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        processed_result = jsonutils.loads(result.body)
        self.assertIn('id', processed_result.keys())
        expected['id'] = processed_result['id']
        self.assertEqual(expected, processed_result)

    def test_get_category(self):
        """Check that get category executed successfully"""

        self._set_policy_rules({'add_category': '@', 'get_category': '@'})
        self.expect_policy_check('add_category')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        expected = {
            'name': 'new_category',
            'created': datetime.isoformat(fake_now)[:-7],
            'updated': datetime.isoformat(fake_now)[:-7],
            'package_count': 0,
            'packages': []
        }

        body = {'name': 'new_category'}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        expected['id'] = jsonutils.loads(result.body)['id']

        self.expect_policy_check('get_category')
        req = self._get('/catalog/categories/%s' % expected['id'])
        retrieved_category = jsonutils.loads(req.get_response(self.api).body)
        self.assertEqual(retrieved_category, expected)

    def test_delete_category(self):
        """Check that category deleted successfully"""
        self._set_policy_rules({'delete_category': '@'})
        self.expect_policy_check('delete_category',
                                 {'category_id': '12345'})

        fake_now = timeutils.utcnow()
        expected = {'name': 'new_category',
                    'created': fake_now,
                    'updated': fake_now,
                    'id': '12345'}

        e = models.Category(**expected)
        test_utils.save_models(e)

        req = self._delete('/catalog/categories/12345')
        processed_result = req.get_response(self.api)
        self.assertEqual(b'', processed_result.body)
        self.assertEqual(200, processed_result.status_code)

    @mock.patch('murano.db.catalog.api.category_get')
    def test_delete_category_with_packages_negative(self, mock_get_category):
        """Check that deleting category that has assigned packages fails."""
        mock_get_category.return_value = mock.MagicMock(
            packages=['test_package'])

        self._set_policy_rules({'delete_category': '@'})
        self.expect_policy_check('delete_category',
                                 {'category_id': '12345'})

        fake_now = timeutils.utcnow()
        expected = {'name': 'new_category',
                    'created': fake_now,
                    'updated': fake_now,
                    'id': '12345'}

        category = models.Category(**expected)
        test_utils.save_models(category)

        req = self._delete('/catalog/categories/12345')
        e = self.assertRaises(exc.HTTPForbidden,
                              self.controller.delete_category, req, '12345')
        self.assertEqual(e.explanation,
                         "It's impossible to delete categories assigned to the"
                         " package, uploaded to the catalog")

    def test_add_category_without_name(self):
        """Test whether adding a category without a name throws exception."""
        self._set_policy_rules({'add_category': '@'})
        self.expect_policy_check('add_category')

        body = {'name': ''}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        e = self.assertRaises(exc.HTTPBadRequest, self.controller.add_category,
                              req, body)
        self.assertEqual("Please, specify a name of the category to create",
                         e.explanation)

    def test_add_category_handle_duplicate_exception(self):
        """Test whether creating duplicate categories throws exception."""
        self._set_policy_rules({'add_category': '@'})
        self.expect_policy_check('add_category')
        self.expect_policy_check('add_category')

        body = {'name': 'new_category'}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        self.controller.add_category(req, body)

        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        e = self.assertRaises(exc.HTTPConflict, self.controller.add_category,
                              req, body)
        self.assertEqual("Category with specified name is already exist",
                         e.explanation)

    def test_add_category_failed_for_non_admin(self):
        """Check that non admin user couldn't add new category"""
        self._set_policy_rules({'add_category': 'role:context_admin'})
        self.is_admin = False
        self.expect_policy_check('add_category')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        body = {'name': 'new_category'}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(403, result.status_code)

    def test_add_long_category(self):
        """Test that category name does not exceed 80 characters

        Check that a category that contains more than 80 characters
        fails to be added
        """

        self._set_policy_rules({'add_category': '@'})
        self.expect_policy_check('add_category')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        body = {'name': 'cat' * 80}
        req = self._post('/catalog/categories', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)
        result_message = result.text.replace('\n', '')
        self.assertIn('Category name should be 80 characters maximum',
                      result_message)

    def test_list_categories(self):
        names = ['cat1', 'cat2']
        for name in names:
            db_catalog_api.category_add(name)

        self._set_policy_rules({'get_category': '@'})
        self.expect_policy_check('get_category')

        req = self._get('/catalog/categories')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        result_categories = jsonutils.loads(result.body)['categories']
        self.assertEqual(2, len(result_categories))
        self.assertEqual(names, [c['name'] for c in result_categories])

        params = {'sort_keys': 'created,  id'}
        req = self._get('/catalog/categories', params)
        self.expect_policy_check('get_category')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        result_categories = jsonutils.loads(result.body)['categories']
        self.assertEqual(names, [c['name'] for c in result_categories])

        names.reverse()

        params = {'sort_dir': 'desc'}
        req = self._get('/catalog/categories', params)
        self.expect_policy_check('get_category')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        result_categories = jsonutils.loads(result.body)['categories']
        self.assertEqual(names, [c['name'] for c in result_categories])

    def test_list_categories_negative(self):
        self._set_policy_rules({'get_category': '@'})
        self.expect_policy_check('get_category')

        req = self._get('/catalog/categories', {'sort_dir': 'test'})
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)

        self.expect_policy_check('get_category')
        req = self._get('/catalog/categories', {'sort_keys': 'test'})
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)

        self.expect_policy_check('get_category')
        req = self._get('/catalog/categories', {'test': ['test']})
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    import congressclient
except ImportError:
    congressclient = None
from unittest import mock

from oslo_config import cfg

from murano.common import engine
from murano.policy import model_policy_enforcer
from murano.tests.unit import base

CONF = cfg.CONF


class TestModelPolicyEnforcer(base.MuranoTestCase):
    def setUp(self):
        if not congressclient:
            self.skipTest('skipped ModelPolicyEnforcer tests')
        super(TestModelPolicyEnforcer, self).setUp()

        self.obj = mock.Mock()
        self.package_loader = mock.Mock()

        self.model_dict = mock.Mock()
        self.obj.to_dictionary = mock.Mock(return_value=self.model_dict)

        self.task = {
            'action': {'method': 'deploy'},
            'model': {'Objects': None,
                      'project_id': 'tenant',
                      'user_id': 'user'
                      },
            'token': 'token',
            'project_id': 'tenant',
            'user_id': 'user',
            'id': 'environment.id'
        }

        self.congress_client_mock = \
            mock.Mock(spec=congressclient.v1.client.Client)
        model_policy_enforcer.ModelPolicyEnforcer._create_client = mock.Mock(
            return_value=self.congress_client_mock)

    def test_enforcer_disabled(self):
        executor = engine.TaskExecutor(self.task)
        executor._model_policy_enforcer = mock.Mock()

        CONF.engine.enable_model_policy_enforcer = False
        executor._validate_model(self.obj, self.package_loader, None)

        self.assertFalse(executor._model_policy_enforcer.validate.called)

    def test_enforcer_enabled(self):
        executor = engine.TaskExecutor(self.task)
        executor._model_policy_enforcer = mock.Mock()

        CONF.engine.enable_model_policy_enforcer = True
        dsl_executor = mock.Mock()
        executor._validate_model(self.obj, self.package_loader, dsl_executor)

        executor._model_policy_enforcer \
            .validate.assert_called_once_with(self.model_dict,
                                              self.package_loader)

    def test_validation_pass(self):
        self.congress_client_mock.execute_policy_action.return_value = \
            {"result": []}
        model = {'?': {'id': '123', 'type': 'class'}}
        enforcer = model_policy_enforcer.ModelPolicyEnforcer(mock.Mock())
        enforcer.validate(model)

    def test_validation_failure(self):
        self.congress_client_mock.execute_policy_action.return_value = \
            {"result": ['predeploy_errors("123","instance1","failure")']}

        model = {'?': {'id': '123', 'type': 'class'}}
        enforcer = model_policy_enforcer.ModelPolicyEnforcer(mock.Mock())
        self.assertRaises(model_policy_enforcer.ValidationError,
                          enforcer.validate, model)

    def test_modify(self):
        model = {'?': {'id': '123', 'type': 'class'}}
        obj = mock.MagicMock()
        obj.to_dictionary = mock.Mock(return_value=model)
        self.congress_client_mock.execute_policy_action.return_value = \
            {"result": [
                'predeploy_modify("123","instance1",'
                '"remove-object: {object_id: "12"}")']}

        action_manager = mock.MagicMock()
        enforcer = model_policy_enforcer.ModelPolicyEnforcer(
            mock.Mock(), action_manager)

        enforcer.modify(obj)
        self.assertTrue(action_manager.apply_action.called)

    def test_parse_result(self):
        congress_response = [
            'unexpected response',
            'predeploy_errors("env1","instance1","Instance 1 has problem")',
            'predeploy_errors("env1","instance1","Instance 2 has problem")',
            'predeploy_errors("env2","instance1","Instance 3 has problem")'
        ]

        enforcer = model_policy_enforcer.ModelPolicyEnforcer(None)
        result = enforcer._parse_simulation_result(
            'predeploy_errors', 'env1', congress_response)

        self.assertNotIn("unexpected response", result)
        self.assertIn("Instance 1 has problem", result)
        self.assertIn("Instance 2 has problem", result)
        self.assertNotIn("Instance 3 has problem", result)

    def test_none_model(self):
        executor = engine.TaskExecutor(self.task)
        executor._model_policy_enforcer = mock.Mock()

        CONF.engine.enable_model_policy_enforcer = True
        dsl_executor = mock.Mock()

        executor._validate_model(None, self.package_loader, dsl_executor)

        self.assertFalse(executor._model_policy_enforcer.modify.called)
        self.assertFalse(executor._model_policy_enforcer.validate.called)

        executor._validate_model(self.obj, self.package_loader, dsl_executor)

        self.assertTrue(executor._model_policy_enforcer.modify.called)
        self.assertTrue(executor._model_policy_enforcer.validate.called)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestReflection(test_case.DslTestCase):
    def setUp(self):
        super(TestReflection, self).setUp()
        self._runner = self.new_runner(om.Object('TestReflection'))

    def test_type_info(self):
        self.assertEqual(
            {
                'name': 'TestReflection',
                'versionStr': '0.0.0',
                'versionMajor': 0,
                'versionMinor': 0,
                'versionPatch': 0,
                'ancestors': ['io.murano.Object'],
                'properties': ['property', 'staticProperty'],
                'package': 'tests',
                'methods': ['foo', 'getAttr', 'setAttr']
            },
            self._runner.testTypeInfo())

    def test_method_info(self):
        self.assertEqual(
            {
                'name': 'foo',
                'arguments': ['bar', 'baz'],
                'barHasDefault': True,
                'bazHasDefault': False,
                'barMethod': 'foo',
                'bazMethod': 'foo',
                'declaringType': 'TestReflection'
            },
            self._runner.testMethodInfo())

    def test_property_info(self):
        self.assertEqual(
            {
                'name': 'property',
                'hasDefault': True,
                'usage': 'InOut'
            },
            self._runner.testPropertyInfo())

    def test_property_read(self):
        self.assertEqual(
            [['object', 'static'], ['static']],
            self._runner.testPropertyRead())

    def test_property_write(self):
        self.assertEqual(
            [['new object', 'new static'], ['new static']],
            self._runner.testPropertyWrite())

    def test_method_invoke(self):
        self.assertEqual('bar baz', self._runner.testMethodInvoke())

    def test_instance_create(self):
        self.assertEqual('test', self._runner.testInstanceCreate())



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import exceptions
from murano.dsl.principal_objects import garbage_collector
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestGC(test_case.DslTestCase):
    def setUp(self):
        super(TestGC, self).setUp()

        self.package_loader.load_package('io.murano', None).register_class(
            garbage_collector.GarbageCollector)
        self.runner = self.new_runner(om.Object('TestGC'))

    def test_model_destroyed(self):
        model = om.Object(
            'TestGCNode', 'root',
            value='root',
            nodes=[
                om.Object(
                    'TestGCNode', 'node1',
                    value='node1',
                    nodes=['root', 'node2']
                ),
                om.Object(
                    'TestGCNode', 'node2',
                    value='node2',
                    nodes=['root', 'node1']
                ),
            ]
        )
        model = {'Objects': None, 'ObjectsCopy': model}
        self.new_runner(model)
        self.assertCountEqual(['node1', 'node2'], self.traces[:2])
        self.assertEqual('root', self.traces[-1])

    def test_collect_from_code(self):
        self.runner.testObjectsCollect()
        self.assertEqual(['B', 'A'], self.traces)

    def test_collect_with_subscription(self):
        self.runner.testObjectsCollectWithSubscription()
        self.assertEqual(
            ['Destroy A', 'Destroy B', 'Destruction of B', 'B', 'A'],
            self.traces)

    def test_call_on_destroyed_object(self):
        self.assertRaises(
            exceptions.ObjectDestroyedError,
            self.runner.testCallOnDestroyedObject)
        self.assertEqual(['foo', 'X'], self.traces)

    def test_destruction_dependencies_serialization(self):
        self.runner.testDestructionDependencySerialization()
        node1 = self.runner.serialized_model['Objects']['outNode']
        node2 = node1['nodes'][0]

        deps = {
            'onDestruction': [{
                'subscriber': self.runner.root.object_id,
                'handler': '_handler'
            }]
        }
        self.assertEqual(deps, node1['?'].get('dependencies'))

        self.assertEqual(
            node1['?'].get('dependencies'),
            node2['?'].get('dependencies'))

        model = self.runner.serialized_model
        model['Objects']['outNode'] = None
        self.new_runner(model)
        self.assertEqual(['Destroy A', 'Destroy B', 'B', 'A'], self.traces)

    def test_is_doomed(self):
        self.runner.testIsDoomed()
        self.assertEqual([[], True, 'B', [True], False, 'A'], self.traces)

    def test_is_destroyed(self):
        self.runner.testIsDestroyed()
        self.assertEqual([False, True], self.traces)

    def test_static_property_not_destroyed(self):
        self.runner.testStaticProperties()
        self.assertEqual([], self.traces)

    def test_args_not_destroyed(self):
        self.runner.testDestroyArgs()
        self.assertEqual([], self.traces)

    def test_runtime_property_not_destroyed(self):
        self.runner.testReachableRuntimeProperties()
        self.assertEqual([False, ], self.traces)



# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import murano.tests.unit.api.base as tb
from unittest import mock

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils

from murano.api.v1 import deployments
from murano.api.v1 import environments

from webob import exc


class TestDeploymentsApi(tb.ControllerTest, tb.MuranoApiTestCase):
    def setUp(self):
        super(TestDeploymentsApi, self).setUp()
        self.environments_controller = environments.Controller()
        self.deployments_controller = deployments.Controller()
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

    def test_deployments_index(self):
        CREDENTIALS = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@',
             'list_deployments': '@'}
        )

        # Create an environment.
        self.expect_policy_check('create_environment')
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS['tenant'], response_body['tenant_id'])
        self.assertEqual('test_environment_1', response_body['name'])
        ENVIRONMENT_ID = response_body['id']

        # Verify that the environment has not yet been deployed.
        self.expect_policy_check('list_deployments',
                                 {'environment_id': ENVIRONMENT_ID})
        result = self.deployments_controller.index(request, ENVIRONMENT_ID)
        self.assertEqual([], result['deployments'])

        # Deploy the environment.
        request = self._post('/environments/{environment_id}/configure'
                             .format(environment_id=ENVIRONMENT_ID),
                             b'', **CREDENTIALS)
        response_body = jsonutils.loads(request.get_response(self.api).body)

        SESSION_ID = response_body['id']

        request = self._post('/environments/{environment_id}/sessions/'
                             '{session_id}/deploy'
                             .format(environment_id=ENVIRONMENT_ID,
                                     session_id=SESSION_ID),
                             b'', **CREDENTIALS)
        result = request.get_response(self.api)
        self.assertEqual('200 OK', result.status)

        # Verify that the environment was deployed.
        self.expect_policy_check('list_deployments',
                                 {'environment_id': ENVIRONMENT_ID})
        result = self.deployments_controller.index(request, ENVIRONMENT_ID)
        deployment_id = result['deployments'][0]['id']
        self.assertEqual(1, len(result['deployments']))
        self.assertIsNotNone(deployment_id)

    def test_deployments_all_environments(self):
        """Test list deployments for all environments.

        Create 2 environments, deploy both, and check that 2 deployments exist.
        """
        CREDENTIALS = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@',
             'list_deployments_all_environments': '@'}
        )

        for count in range(2):
            # Create environment.
            self.expect_policy_check('create_environment')
            request = self._post(
                '/environments',
                jsonutils.dump_as_bytes(
                    {'name': 'test_environment_{0}'.format(count)}),
                **CREDENTIALS
            )
            response_body = jsonutils.loads(
                request.get_response(self.api).body)
            self.assertEqual(CREDENTIALS['tenant'], response_body['tenant_id'])
            self.assertEqual('test_environment_{0}'.format(count),
                             response_body['name'])
            ENVIRONMENT_ID = response_body['id']

            # Deploy environment.
            request = self._post('/environments/{environment_id}/configure'
                                 .format(environment_id=ENVIRONMENT_ID),
                                 b'', **CREDENTIALS)
            response_body = jsonutils.loads(
                request.get_response(self.api).body)
            SESSION_ID = response_body['id']
            request = self._post('/environments/{environment_id}/sessions/'
                                 '{session_id}/deploy'
                                 .format(environment_id=ENVIRONMENT_ID,
                                         session_id=SESSION_ID),
                                 b'', **CREDENTIALS)
            result = request.get_response(self.api)
            self.assertEqual('200 OK', result.status)

        # Check that 2 deployments exist.
        self.expect_policy_check('list_deployments_all_environments')
        result = self.deployments_controller.index(request, None)
        self.assertEqual(2, len(result['deployments']))
        for deployment in result['deployments']:
            self.assertIsNotNone(deployment)
        self.assertNotEqual(result['deployments'][0], result['deployments'][1])

    def test_deployments_all_environments_different_tenants(self):
        """Test list deployments for all environments in different tenants.

        Should only return return environments for current tenant.
        """
        CREDENTIALS = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        ALT_CREDENTIALS = {'tenant': 'test_tenant_2', 'user': 'test_user_2'}
        self._set_policy_rules(
            {'create_environment': '@',
             'list_deployments_all_environments': '@'}
        )

        deployments = []

        # Create the first environment inside first tenant and the second
        # environments inside the alternate tenant. Then deploy both.
        for count, creds in enumerate([CREDENTIALS, ALT_CREDENTIALS]):
            # Create each environment.
            self.expect_policy_check('create_environment')
            request = self._post(
                '/environments',
                jsonutils.dump_as_bytes(
                    {'name': 'test_environment_{0}'.format(count)}),
                **creds
            )
            response_body = jsonutils.loads(
                request.get_response(self.api).body)
            self.assertEqual(creds['tenant'], response_body['tenant_id'])
            self.assertEqual('test_environment_{0}'.format(count),
                             response_body['name'])
            ENVIRONMENT_ID = response_body['id']

            # Deploy each environment.
            request = self._post('/environments/{environment_id}/configure'
                                 .format(environment_id=ENVIRONMENT_ID),
                                 b'', **creds)
            response_body = jsonutils.loads(
                request.get_response(self.api).body)
            SESSION_ID = response_body['id']
            request = self._post('/environments/{environment_id}/sessions/'
                                 '{session_id}/deploy'
                                 .format(environment_id=ENVIRONMENT_ID,
                                         session_id=SESSION_ID),
                                 b'', **creds)
            result = request.get_response(self.api)
            self.assertEqual('200 OK', result.status)

            # Check that each tenant only returns one deployment.
            self.expect_policy_check('list_deployments_all_environments')
            result = self.deployments_controller.index(request, None)
            self.assertEqual(1, len(result['deployments']))
            deployment_id = result['deployments'][0]['id']
            self.assertIsNotNone(deployment_id)
            deployments.append(deployment_id)

        self.assertNotEqual(deployments[0], deployments[1])

    def test_deployments_not_found_statuses(self):
        CREDENTIALS = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@',
             'statuses_deployments': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post('/environments', jsonutils.
                             dump_as_bytes({'name': 'test_environment_1'}),
                             **CREDENTIALS)
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS['tenant'],
                         response_body['tenant_id'])

        ENVIRONMENT_ID = response_body['id']
        deploy_id = '1'
        self.expect_policy_check('statuses_deployments',
                                 {'deployment_id': deploy_id,
                                  'environment_id': ENVIRONMENT_ID})
        self.assertRaises(exc.HTTPNotFound,
                          self.deployments_controller.statuses, request,
                          ENVIRONMENT_ID, deploy_id)

    def test_deployments_status(self):
        CREDENTIALS = {'tenant': 'test_tenant', 'user': 'test_user'}
        self._set_policy_rules(
            {'create_environment': '@',
             'statuses_deployments': '@',
             'list_deployments': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment'}),
            **CREDENTIALS
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS['tenant'],
                         response_body['tenant_id'])

        ENVIRONMENT_ID = response_body['id']

        # Create session
        request = self._post('/environments/{environment_id}/configure'
                             .format(environment_id=ENVIRONMENT_ID),
                             b'', **CREDENTIALS)
        response_body = jsonutils.loads(request.get_response(self.api).body)

        SESSION_ID = response_body['id']

        request = self._post('/environments/{environment_id}/sessions/'
                             '{session_id}/deploy'
                             .format(environment_id=ENVIRONMENT_ID,
                                     session_id=SESSION_ID),
                             b'', **CREDENTIALS)
        request.get_response(self.api)

        self.expect_policy_check('list_deployments',
                                 {'environment_id': ENVIRONMENT_ID})
        result = self.deployments_controller.index(request, ENVIRONMENT_ID)
        deploy_id = result['deployments'][0]['id']

        self.expect_policy_check('statuses_deployments',
                                 {'deployment_id': deploy_id,
                                  'environment_id': ENVIRONMENT_ID})
        result = self.deployments_controller.statuses(request, ENVIRONMENT_ID,
                                                      deploy_id)
        self.assertNotEqual(result['reports'], [])

        request.GET['service_id'] = "12"
        self.expect_policy_check('statuses_deployments',
                                 {'deployment_id': deploy_id,
                                  'environment_id': ENVIRONMENT_ID})
        result = self.deployments_controller.statuses(request, ENVIRONMENT_ID,
                                                      deploy_id)
        self.assertEqual(result['reports'], [])

        self.expect_policy_check('statuses_deployments',
                                 {'deployment_id': deploy_id,
                                  'environment_id': '12'})
        self.assertRaises(exc.HTTPBadRequest,
                          self.deployments_controller.statuses, request,
                          "12", deploy_id)

    def test_set_dep_state(self):
        deployment = mock.Mock()
        deployment.id = "1"
        deployment.description = {'applications': []}
        unit = mock.Mock()
        query_result = mock.Mock()
        filter_by_result = mock.Mock()
        filter_by_result.count.return_value = 0
        query_result.filter_by.return_value = filter_by_result
        unit.query.return_value = query_result

        result = deployments.set_dep_state(deployment, unit)
        self.assertEqual('success', result.state)

        deployment.description = None
        result = deployments.set_dep_state(deployment, unit)
        self.assertEqual('success', result.state)

        filter_by_result.count.return_value = 1
        query_result.filter_by.return_value = filter_by_result
        result = deployments.set_dep_state(deployment, unit)
        self.assertEqual('completed_w_errors', result.state)

        filter_by_result.count.return_value = 0
        query_result.filter_by.return_value = filter_by_result
        deployment.finished = False
        result = deployments.set_dep_state(deployment, unit)
        self.assertEqual('running', result.state)

        filter_by_result.count.return_value = 1
        query_result.filter_by.return_value = filter_by_result
        result = deployments.set_dep_state(deployment, unit)
        self.assertEqual('running_w_errors', result.state)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

RULE_ADMIN_API = 'rule:admin_api'
RULE_DEFAULT = 'rule:default'


rules = [
    policy.RuleDefault(
        name='context_is_admin',
        check_str='role:admin'),
    policy.RuleDefault(
        name='admin_api',
        check_str='is_admin:True'),
    policy.RuleDefault(
        name='default',
        check_str='')
]


def list_rules():
    return rules



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for service broker data
"""
from oslo_db.sqlalchemy import models
import sqlalchemy as sa
from sqlalchemy.ext import declarative


class _ServiceBrokerBase(models.ModelBase):
    pass


Base = declarative.declarative_base(cls=_ServiceBrokerBase)


class CFOrganization(Base):
    __tablename__ = "cf_orgs"
    id = sa.Column(sa.String(255), primary_key=True)
    tenant = sa.Column(sa.String(255), nullable=False)


class CFSpace(Base):
    __tablename__ = "cf_spaces"

    id = sa.Column(sa.String(255), primary_key=True)
    environment_id = sa.Column(sa.String(255), nullable=False)


class CFServiceInstance(Base):
    __tablename__ = 'cf_serv_inst'

    id = sa.Column(sa.String(255), primary_key=True)
    service_id = sa.Column(sa.String(255), nullable=False)
    environment_id = sa.Column(sa.String(255), nullable=False)
    tenant = sa.Column(sa.String(255), nullable=False)


def register_models(engine):
    """Creates database tables for all models with the given engine."""
    models = (CFSpace, CFOrganization, CFServiceInstance)
    for model in models:
        model.metadata.create_all(engine)


def unregister_models(engine):
    """Drops database tables for all models with the given engine."""
    models = (CFOrganization, CFSpace, CFServiceInstance)
    for model in models:
        model.metadata.drop_all(engine)



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import yaml

from murano.packages import exceptions
import murano.packages.mpl_package as mpl_package
import murano.tests.unit.base as test_base


class TestMPLPackage(test_base.MuranoTestCase):

    def setUp(cls):
        super(TestMPLPackage, cls).setUp()
        cls.source_directory = os.path.dirname(os.path.realpath(__file__))
        manifest_path = os.path.join(cls.source_directory, 'manifest.yaml')
        cls.manifest = {}
        with open(manifest_path) as manifest_file:
            for key, value in yaml.safe_load(manifest_file).items():
                cls.manifest[key] = value

    def test_classes_property(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        classes = package.classes
        self.assertIn('Class1', classes)
        self.assertIn('Class2', classes)

    def test_ui_property(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        ui = package.ui
        self.assertIsNotNone(ui)
        self.assertIn(b'name: license', ui)
        self.assertIn(b'type: string', ui)
        self.assertIn(b'description: Apache License, Version 2.0', ui)
        self.assertIn(b'hidden: false', ui)
        self.assertIn(b'required: false', ui)

    def test_requirements_property(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        requirements = package.requirements
        self.assertEqual(requirements, {})

        # Override the Require property in cls.manifest and check new value.
        self.manifest['Require'] = {'murano.plugins.example': 0}
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        requirements = package.requirements
        self.assertEqual(requirements, {'murano.plugins.example': 0})
        del self.manifest['Require']

    def test_meta_property(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        meta = package.meta
        self.assertEqual(meta, 'test.meta')

    def test_get_class(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        stream, path = package.get_class('Class1')
        expected_path = os.path.join(self.source_directory, 'Classes',
                                     'test.class1')
        self.assertIn(b'test.class1', stream)
        self.assertEqual(path, expected_path)

    def test_get_class_with_inappropriate_name(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        self.assertRaises(exceptions.PackageClassLoadError,
                          package.get_class,
                          'Invalid name')

    def test_get_class_with_nonexistent_class(self):
        package = mpl_package.MuranoPlPackage(None, None,
                                              self.source_directory,
                                              self.manifest)
        self.assertRaises(exceptions.PackageClassLoadError,
                          package.get_class,
                          'Class2')




#    Copyright (c) 2015 Telefonica I+D.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools as func

from oslo_log import log as logging
from webob import exc

from murano.api.v1 import request_statistics
from murano.common.helpers import token_sanitizer
from murano.common.i18n import _
from murano.common import policy
from murano.common import wsgi
from murano.db.services import core_services
from murano import utils


LOG = logging.getLogger(__name__)

API_NAME = 'Services'


def normalize_path(f):
    """Normalizes request path

       It normalizes the path obtaining in the requests.
       It is used in all the operations in the controller
    """
    @func.wraps(f)
    def f_normalize_path(*args, **kwargs):
        if 'path' in kwargs:
            if kwargs['path']:
                kwargs['path'] = '/services/' + kwargs['path']
            else:
                kwargs['path'] = '/services'
        return f(*args, **kwargs)

    return f_normalize_path


class Controller(object):
    @request_statistics.stats_count(API_NAME, 'Index')
    @utils.verify_env_template
    @normalize_path
    def index(self, request, env_template_id, path):
        """Obtains services/applications for a template

           It obtains all the services/applications associated to
           a template
           :param request: The operation request.
           :param env_template_id: The environment template id with
           contains the services
           :param path: The operation path
        """
        LOG.debug('Applications:Get <EnvTemplateId: {templ_id}, '
                  'Path: {path}>'.format(templ_id=env_template_id, path=path))

        try:
            get_data = core_services.CoreServices.get_template_data
            result = get_data(env_template_id, path)
        except (KeyError, ValueError, AttributeError):
            msg = _('The environment template {templ_id} does not '
                    'exist').format(templ_id=env_template_id)
            LOG.exception(msg)
            raise exc.HTTPNotFound(msg)
        return result

    @request_statistics.stats_count(API_NAME, 'Show')
    @utils.verify_env_template
    @normalize_path
    def show(self, request, env_template_id, path):
        """It shows the service description

        :param request: The operation request.
        :param env_template_id: the env template ID where the service
        belongs to.
        :param path: The path include the service id
        :return: the service description.
        """
        LOG.debug('Applications:Get <EnvTemplateId: {templ_id}, '
                  'Path: {path}>'.format(templ_id=env_template_id, path=path))

        try:
            get_data = core_services.CoreServices.get_template_data
            result = get_data(env_template_id, path)
        except (KeyError, ValueError, AttributeError):
            msg = _('The template does not exist {templ_id}').format(
                templ_id=env_template_id)
            LOG.exception(msg)
            raise exc.HTTPNotFound(msg)
        return result

    @request_statistics.stats_count(API_NAME, 'Create')
    @utils.verify_env_template
    @normalize_path
    def post(self, request, env_template_id, path, body):
        """It adds a service into a template

        :param request: The operation request.
        :param env_template_id: the env template ID where the
        service belongs to.
        :param path: The path
        :param body: the information about the service
        :return: the service description.
        """
        secure_data = token_sanitizer.TokenSanitizer().sanitize(body)
        LOG.debug('Applications:Post <EnvTempId: {env_id}, Path: {path}, '
                  'Body: {body}>'.format(env_id=env_template_id,
                                         body=secure_data,
                                         path=path))

        post_data = core_services.CoreServices.post_application_data
        try:
            result = post_data(env_template_id, body, path)
        except (KeyError, ValueError):
            msg = _('The template does not exist {templ_id}').format(
                templ_id=env_template_id)
            LOG.exception(msg)
            raise exc.HTTPNotFound(msg)
        return result

    @request_statistics.stats_count(API_NAME, 'Update')
    @utils.verify_env_template
    @normalize_path
    def put(self, request, env_template_id, path, body):

        """It updates a service into a template.

        :param request: The operation request.
        :param env_template_id: the env template ID where the service
        belongs to.
        :param path: The path
        :param body: the information about the service
        :return: the service description updated.
        """
        policy.check('update_service_env_template', request.context)
        LOG.debug('Applications:Put <EnvTempId: {templ_id}, Path: {path}, '
                  'Body: {body}>'.format(templ_id=env_template_id,
                                         body=body,
                                         path=path))

        put_data = core_services.CoreServices.put_application_data

        try:
            result = put_data(env_template_id, body, path)
        except (KeyError, ValueError):
            msg = _('The template does not exist {templ_id}').format(
                templ_id=env_template_id)
            LOG.exception(msg)
            raise exc.HTTPNotFound(msg)
        return result

    @request_statistics.stats_count(API_NAME, 'Delete')
    @utils.verify_env_template
    @normalize_path
    def delete(self, request, env_template_id, path):
        """It deletes a service into a template.

        :param request: The operation request.
        :param env_template_id: the env template ID where
        the service belongs to.
        :param path: The path contains the service id
        """
        LOG.debug('Applications:Put <EnvTempId: {templ_id}, '
                  'Path: {path}>'.format(templ_id=env_template_id,
                                         path=path))
        delete_data = core_services.CoreServices.delete_env_template_data
        try:
            result = delete_data(env_template_id, path)
        except (KeyError, ValueError):
            msg = _('The template does not exist {templ_id}').format(
                templ_id=env_template_id)
            LOG.exception(msg)
            raise exc.HTTPNotFound(msg)
        return result


def create_resource():
    return wsgi.Resource(Controller())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import semantic_version
import types
from unittest import mock
import weakref

from oslo_utils.uuidutils import generate_uuid

from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.dsl import helpers
from murano.tests.unit import base


class TestDSLHelpers(base.MuranoTestCase):
    @mock.patch.object(helpers, 'with_object_store', autospec=True)
    def test_parallel_select_except_exception(self, mock_with_object_store):
        mock_with_object_store.side_effect = ValueError
        self.assertRaises(ValueError, helpers.parallel_select,
                          [mock.sentinel.foo], lambda: None)

    def test_enum(self):
        self.assertEqual('Enum', helpers.enum().__name__)

    def test_cast_with_murano_type(self):
        mock_attrs = {
            'name': mock.sentinel.class_type,
            'version': semantic_version.Version('1.0.0'),
            'ancestors.return_value': []
        }
        mock_type = mock.Mock()
        mock_type.configure_mock(**mock_attrs)
        mock_obj = mock.Mock(type=mock_type)
        mock_obj.cast.return_value = mock.sentinel.foo_cast_value
        mock_murano_class = mock.Mock(spec=dsl_types.MuranoType)
        mock_murano_class.name = mock.sentinel.class_type
        mock_murano_class.version = semantic_version.Version('1.0.0')

        result = helpers.cast(mock_obj, mock_murano_class,
                              pov_or_version_spec=None)
        self.assertEqual(mock.sentinel.foo_cast_value, result)
        mock_obj.cast.assert_called_once_with(mock_type)

    def test_cast_except_value_error(self):
        mock_attrs = {
            'name': mock.sentinel.class_type,
            'version': semantic_version.Version('1.0.0'),
            'ancestors.return_value': []
        }
        mock_type = mock.Mock()
        mock_type.configure_mock(**mock_attrs)
        mock_obj = mock.Mock(type=mock_type)
        mock_murano_class = mock.Mock(spec=dsl_types.MuranoType)
        mock_murano_class.name = mock.sentinel.class_type

        e = self.assertRaises(ValueError, helpers.cast, mock_obj,
                              mock_murano_class,
                              pov_or_version_spec=mock.Mock())
        self.assertEqual('pov_or_version_spec of unsupported type {0}'
                         .format(type(mock.Mock())), str(e))

    def test_cast_except_no_class_found(self):
        mock_attrs = {
            'name': mock.sentinel.name,
            'package.name': mock.sentinel.package_name,
            'version': mock.sentinel.version,
            'ancestors.return_value': []
        }
        mock_type = mock.Mock()
        mock_type.configure_mock(**mock_attrs)
        mock_obj = mock.Mock(type=mock_type)
        mock_murano_class = mock.Mock(spec=dsl_types.MuranoTypeReference)
        mock_murano_class.type = mock.sentinel.foo_class
        mock_version_spec = mock.Mock(spec=dsl_types.MuranoPackage)

        e = self.assertRaises(exceptions.NoClassFound, helpers.cast, mock_obj,
                              mock_murano_class,
                              pov_or_version_spec=mock_version_spec)
        self.assertIn('Class "sentinel.foo_class" is not found', str(e))

    def test_cast_except_ambiguous_class_name(self):
        mock_attrs = {
            'name': mock.sentinel.class_type,
            'version': semantic_version.Version('1.0.0')
        }
        mock_ancestor = mock.Mock()
        mock_ancestor.configure_mock(**mock_attrs)
        mock_attrs['ancestors.return_value'] = [mock_ancestor]
        mock_type = mock.Mock()
        mock_type.configure_mock(**mock_attrs)
        mock_obj = mock.Mock(type=mock_type)
        mock_murano_class = mock.Mock(spec=dsl_types.MuranoTypeReference)
        mock_murano_class.type = mock.sentinel.class_type

        # pov_or_version_spec of '1' will be converted to
        # semantic_version.Spec('>=1.0.0,<2.0.0-0')
        self.assertRaises(exceptions.AmbiguousClassName, helpers.cast,
                          mock_obj, mock_murano_class, pov_or_version_spec='1')

    def test_inspect_is_method(self):
        mock_cls = mock.Mock(foo=lambda: None, bar=None)
        self.assertTrue(helpers.inspect_is_method(mock_cls, 'foo'))
        self.assertFalse(helpers.inspect_is_method(mock_cls, 'bar'))

    def test_inspect_is_property(self):
        data_descriptor = mock.MagicMock(__get__=None, __set__=None)
        mock_cls = mock.Mock(foo=data_descriptor, bar=None)
        self.assertTrue(helpers.inspect_is_property(mock_cls, 'foo'))
        self.assertFalse(helpers.inspect_is_property(mock_cls, 'bar'))

    def test_updated_dict(self):
        dict_ = {'foo': 'bar'}
        self.assertEqual(dict_, helpers.updated_dict(dict_, {}))

    def test_updated_dict_with_null_arg(self):
        dict_ = {'foo': 'bar'}
        self.assertEqual(dict_, helpers.updated_dict(None, dict_))

    def test_resolve_with_return_reference_true(self):
        mock_value = mock.Mock(spec=dsl_types.MuranoTypeReference)
        mock_scope_type = mock.Mock(spec=dsl_types.MuranoTypeReference)
        result = helpers.resolve_type(mock_value, mock_scope_type, True)
        self.assertEqual(mock_value, result)

        mock_value = mock.Mock()
        mock_value.get_reference.return_value = mock.sentinel.foo_reference
        mock_scope_type = mock.Mock()
        mock_scope_type.package.find_class.return_value = mock_value
        result = helpers.resolve_type(mock_value, mock_scope_type, True)
        self.assertEqual(mock.sentinel.foo_reference, result)

    def test_resolve_type_with_null_value(self):
        self.assertIsNone(helpers.resolve_type(None, None))

    def test_assemble_object_definition(self):
        test_parsed = {
            'type': mock.sentinel.type,
            'properties': {},
            'id': mock.sentinel.id,
            'name': mock.sentinel.name,
            'metadata': mock.sentinel.metadata,
            'destroyed': True,
            'extra': {}
        }
        expected = {
            '?': {
                'type': mock.sentinel.type,
                'id': mock.sentinel.id,
                'name': mock.sentinel.name,
                'metadata': mock.sentinel.metadata,
                'destroyed': True
            }
        }

        result = helpers.assemble_object_definition(test_parsed)
        for key, val in expected.items():
            self.assertEqual(val, result[key])

    @mock.patch.object(helpers, 'format_type_string', autospec=True)
    def test_assemble_object_definition_with_serializable_model_format(
            self, mock_format_type_string):
        mock_format_type_string.return_value = mock.sentinel.type

        test_parsed = {
            'type': mock.sentinel.type,
            'properties': {},
            'id': mock.sentinel.id,
            'name': mock.sentinel.name,
            'metadata': mock.sentinel.metadata,
            'destroyed': True,
            'extra': {}
        }
        expected = {
            '?': {
                'type': mock.sentinel.type,
                'id': mock.sentinel.id,
                'name': mock.sentinel.name,
                'metadata': mock.sentinel.metadata,
                'destroyed': True
            }
        }
        model_format = dsl_types.DumpTypes.Serializable

        result = helpers.assemble_object_definition(test_parsed, model_format)
        for key, val in expected['?'].items():
            self.assertEqual(val, result['?'][key])

    def test_assemble_object_definition_with_inline_model_format(self):
        test_parsed = {
            'type': mock.sentinel.type,
            'properties': mock.sentinel.properties,
            'id': mock.sentinel.id,
            'name': mock.sentinel.name,
            'metadata': mock.sentinel.metadata,
            'dependencies': mock.sentinel.dependencies,
            'destroyed': mock.sentinel.destroyed,
            'extra': {}
        }
        model_format = dsl_types.DumpTypes.Inline
        expected = copy.copy(test_parsed)
        expected[mock.sentinel.type] = mock.sentinel.properties
        for key in ['type', 'extra', 'properties']:
            expected.pop(key)

        result = helpers.assemble_object_definition(test_parsed, model_format)
        for key, val in expected.items():
            self.assertEqual(val, result[key])

    def test_assemble_object_definition_except_value_error(self):
        test_parsed = {
            'type': mock.sentinel.type,
            'properties': {},
            'id': mock.sentinel.id,
            'name': mock.sentinel.name,
            'metadata': mock.sentinel.metadata,
            'destroyed': True,
            'extra': {}
        }
        e = self.assertRaises(ValueError, helpers.assemble_object_definition,
                              test_parsed, None)
        self.assertEqual('Invalid Serialization Type', str(e))

    def test_function(self):
        def f():
            return
        self.assertTrue(isinstance(helpers.function(f), types.FunctionType))

    def test_function_from_method(self):
        class C:
            def m(self):
                return
        c = C()
        self.assertTrue(isinstance(helpers.function(c.m), types.FunctionType))

    def test_weak_proxy(self):
        self.assertIsNone(helpers.weak_proxy(None))

    def test_weak_proxy_with_reference_type(self):
        result = helpers.weak_proxy(weakref.ReferenceType(int))
        self.assertEqual('int', result.__name__)

    @mock.patch.object(helpers, 'get_object_store', autospec=True)
    def test_weak_ref(self, mock_get_object_store):
        mock_object_store = mock.Mock(
            **{'get.return_value': mock.sentinel.res})
        mock_get_object_store.return_value = mock_object_store

        test_obj = dsl_types.MuranoObject()
        setattr(test_obj, 'object_id', generate_uuid())
        murano_object_weak_ref = helpers.weak_ref(test_obj)
        setattr(murano_object_weak_ref, 'ref', lambda *args: None)
        result = murano_object_weak_ref.__call__()

        self.assertEqual(mock.sentinel.res, result)
        self.assertEqual('weakref',
                         murano_object_weak_ref.ref.__class__.__name__)

    def test_weak_ref_with_null_obj(self):
        self.assertIsNone(helpers.weak_ref(None))

    @mock.patch.object(helpers, 're', autospec=True)
    def test_parse_type_string_with_null_res(self, mock_re):
        mock_re.compile.return_value = mock.Mock(
            **{'match.return_value': None})
        self.assertIsNone(helpers.parse_type_string('', None, None))

    def test_format_type_string(self):
        inner_type_obj = mock.Mock(spec=dsl_types.MuranoType)
        inner_type_obj.configure_mock(**{'name': 'foo', 'version': 'foo_ver'})
        inner_type_obj_pkg = mock.Mock()
        inner_type_obj_pkg.configure_mock(name='foo_pkg')
        setattr(inner_type_obj, 'package', inner_type_obj_pkg)
        type_obj = mock.Mock(spec=dsl_types.MuranoTypeReference,
                             type=inner_type_obj)
        result = helpers.format_type_string(type_obj)
        self.assertEqual('foo/foo_ver@foo_pkg', result)

    def test_format_type_string_except_value_error(self):
        type_obj = mock.Mock(spec=dsl_types.MuranoTypeReference, type=None)
        e = self.assertRaises(ValueError, helpers.format_type_string, type_obj)
        self.assertEqual('Invalid argument', str(e))

    def test_patch_dict(self):
        path = 'foo.bar.baz'
        fake_dict = mock.MagicMock(spec=dict)
        # Make the dict return itself to test whether all the parts are called.
        fake_dict.get.return_value = fake_dict
        helpers.patch_dict(fake_dict, path, None)
        fake_dict.get.assert_has_calls([mock.call('foo'), mock.call('bar')])
        fake_dict.pop.assert_not_called()

    def test_patch_dict_without_dict(self):
        path = 'foo.bar.baz'
        not_a_dict = mock.Mock()
        helpers.patch_dict(not_a_dict, path, None)
        not_a_dict.get.assert_not_called()
        not_a_dict.pop.assert_not_called()

    @mock.patch.object(helpers, 'gc')
    def test_walk_gc_with_towards_true(self, mock_gc, autospec=True):
        mock_gc.get_referrers.side_effect = [
            [mock.sentinel.second], [mock.sentinel.third]
        ]
        first_obj = mock.sentinel.first
        handler = mock.MagicMock()
        handler.return_value = True

        expected = [
            [mock.sentinel.first],
            [mock.sentinel.first, mock.sentinel.second],
            [mock.sentinel.first, mock.sentinel.second, mock.sentinel.third]
        ]
        actual = []
        for obj in helpers.walk_gc(first_obj, True, handler):
            actual.append(obj)
        self.assertEqual(expected, actual)

    @mock.patch.object(helpers, 'gc', autospec=True)
    def test_walk_gc_with_towards_false(self, mock_gc):
        mock_gc.get_referents.side_effect = [
            # Trigger the continue by duplicating entries.
            [mock.sentinel.second], [mock.sentinel.second]
        ]
        first_obj = mock.sentinel.first
        handler = mock.MagicMock()
        handler.return_value = True

        expected = [
            [mock.sentinel.first],
            [mock.sentinel.second, mock.sentinel.first]
        ]
        actual = []
        for obj in helpers.walk_gc(first_obj, False, handler):
            actual.append(obj)
        self.assertEqual(expected, actual)


class TestMergeDicts(base.MuranoTestCase):
    def check(self, dict1, dict2, expected):
        result = helpers.merge_dicts(dict1, dict2)
        self.assertEqual(expected, result)

    def test_dicts_plain(self):
        dict1 = {"a": "1"}
        dict2 = {"a": "100", "ab": "12"}
        expected = {"a": "100", "ab": "12"}
        self.check(dict1, dict2, expected)

    def test_different_types_none(self):
        dict1 = {"a": "1"}
        dict2 = {"a": None, "ab": "12"}
        expected = {"a": "1", "ab": "12"}
        self.check(dict1, dict2, expected)

    def test_different_types_of_iterable(self):
        dict1 = {"a": {"ab": "1"}}
        dict2 = {"a": ["ab", "1"]}
        self.assertRaises(TypeError, helpers.merge_dicts, dict1, dict2)

    def test_merge_nested_dicts(self):
        dict1 = {"a": {"ab": {}, "abc": "1"}}
        dict2 = {"a": {"abc": "123"}}
        expected = {"a": {"ab": {}, "abc": "123"}}
        self.check(dict1, dict2, expected)

    def test_merge_nested_dicts_with_max_levels(self):
        dict1 = {"a": {"ab": {"abcd": "1234"}, "abc": "1"}}
        dict2 = {"a": {"ab": {"y": "9"}, "abc": "123"}}
        expected = {"a": {"ab": {"y": "9"}, "abc": "123"}}
        result = helpers.merge_dicts(dict1, dict2, max_levels=2)
        self.assertEqual(expected, result)

    def test_merge_with_lists(self):
        dict1 = {"a": [1, 2]}
        dict2 = {"a": [1, 3, 2, 4]}
        expected = {"a": [1, 2, 3, 4]}
        self.check(dict1, dict2, expected)


class TestParseVersionSpec(base.MuranoTestCase):
    def check(self, expected, version_spec):
        self.assertEqual(expected, helpers.parse_version_spec(version_spec))

    def test_empty_version_spec(self):
        version_spec = ""
        expected = semantic_version.Spec('>=0.0.0', '<1.0.0')
        self.check(expected, version_spec)

    def test_empty_kind(self):
        version_spec = "1.11.111"
        expected = semantic_version.Spec('==1.11.111')
        self.check(expected, version_spec)

    def test_implicit_major(self):
        version_spec = ">=2"
        expected = semantic_version.Spec('>=2.0.0')
        self.check(expected, version_spec)

    def test_implicit_minor(self):
        version_spec = ">=2.1"
        expected = semantic_version.Spec('>=2.1.0')
        self.check(expected, version_spec)

    def test_remove_spaces(self):
        version_spec = "< =  2 .1"
        expected = semantic_version.Spec('<2.2.0')
        self.check(expected, version_spec)

    def test_input_version(self):
        version_spec = semantic_version.Version('1.11.111')
        expected = semantic_version.Spec('==1.11.111')
        self.check(expected, version_spec)

    def test_input_spec(self):
        version_spec = semantic_version.Spec('<=1', '<=1.11')
        expected = semantic_version.Spec('<1.12.0', '<2.0.0')
        self.check(expected, version_spec)



# Copyright (c) 2016 AT&T inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import eventlet
import socket
from unittest import mock

import webob

from oslo_config import cfg
from xml.dom import minidom

from murano.common import exceptions
from murano.common.i18n import _
from murano.common import wsgi
from murano.tests.unit import base

CONF = cfg.CONF


class TestServiceBrokerResponseSerializer(base.MuranoTestCase):

    def test_service_broker_response_serializer(self):
        self.sbrs = wsgi.ServiceBrokerResponseSerializer()
        result = self.sbrs.serialize("test_response", "application/json")
        self.assertEqual(200, result.status_code)

        response = webob.Response("test_body")
        response.data = "test_data"
        result = self.sbrs.serialize(response, "application/json")
        self.assertEqual(200, result.status_code)


class TestResponseSerializer(base.MuranoTestCase):

    def test_response_serializer(self):
        self.response_serializer = wsgi.ResponseSerializer(None)
        self.assertRaises(exceptions.UnsupportedContentType,
                          self.response_serializer.get_body_serializer,
                          "not_valid")


class TestRequestDeserializer(base.MuranoTestCase):

    def test_request_deserializer_deserialize_body(self):
        self.request_deserializer = wsgi.RequestDeserializer()
        request = mock.Mock()
        request.get_content_type.side_effect = (
            exceptions.UnsupportedContentType
        )
        request.body = "body"
        self.assertRaises(exceptions.UnsupportedContentType,
                          self.request_deserializer.deserialize_body,
                          request, "act")

        request.get_content_type.side_effect = None
        request.get_content_type.return_value = None
        result = self.request_deserializer.deserialize_body(request, "act")
        self.assertEqual({}, result)

        request.get_content_type.return_value = ""
        self.assertRaises(exceptions.UnsupportedContentType,
                          self.request_deserializer.deserialize_body,
                          request, "act")

    def test_request_deserializer_get_action_args(self):
        self.request_deserializer = wsgi.RequestDeserializer()
        request_environment = []
        result = self.request_deserializer.get_action_args(request_environment)
        self.assertEqual({}, result)

        request_environment = {'wsgiorg.routing_args': ["", {}]}
        result = self.request_deserializer.get_action_args(request_environment)
        self.assertEqual({}, result)


class TestService(base.MuranoTestCase):

    def setUp(self):
        super(TestService, self).setUp()
        # Greatly speed up the time it takes to run the tests that call
        # wsgi._get_socket below: retry_until will be set to 31, the 1st
        # iteration of the while loop will execute because it will evaluate to
        # 30 < 31, but 2nd one will not because it will evaluate to 31 < 31.
        self.mock_time = mock.patch.object(
            wsgi, 'time', **{'time.side_effect': [1, 30, 31]}).start()
        self.addCleanup(mock.patch.stopall)

    @mock.patch("murano.common.wsgi.socket")
    def test_service_get_socket(self, socket):
        self.service = wsgi.Service(None, 1)
        new_socket = self.service._get_socket(None, None, None)
        self.assertIsInstance(new_socket, eventlet.greenio.base.GreenSocket)

        self.mock_time.time.side_effect = [1, 30, 31]
        socket.TCP_KEEPIDLE = True
        new_socket_2 = self.service._get_socket(None, None, None)
        self.assertIsInstance(new_socket_2, eventlet.greenio.base.GreenSocket)

    @mock.patch("murano.common.wsgi.socket")
    @mock.patch("murano.common.wsgi.sslutils")
    def test_service_get_socket_sslutils_enabled(self, sslutils, mock_socket):
        self.service = wsgi.Service(None, 1)
        new_socket = self.service._get_socket(None, None, None)
        self.assertIsNotNone(new_socket)

    @mock.patch("murano.common.wsgi.socket")
    @mock.patch("murano.common.wsgi.eventlet")
    def test_service_get_socket_no_bind(self, eventlet, mock_socket):
        self.service = wsgi.Service(None, 1)
        eventlet.listen.return_value = None
        self.assertRaises(RuntimeError, self.service._get_socket,
                          None, None, None)

    @mock.patch("murano.common.wsgi.socket")
    @mock.patch("murano.common.wsgi.eventlet")
    def test_service_get_socket_os_error(self, eventlet, mock_socket):
        mock_socket.error = socket.error
        self.service = wsgi.Service(None, 1)
        sock_err = socket.error(1)
        eventlet.listen.side_effect = sock_err
        self.assertRaises(socket.error, self.service._get_socket,
                          None, None, None)

    @mock.patch("murano.common.wsgi.socket")
    @mock.patch("murano.common.wsgi.eventlet")
    def test_service_get_socket_socket_error_EADDRINUSE(self, eventlet,
                                                        mock_socket):
        mock_socket.error = socket.error
        self.service = wsgi.Service(None, 1)
        sock_err = socket.error(errno.EADDRINUSE)
        eventlet.listen.side_effect = sock_err
        self.assertRaises(RuntimeError, self.service._get_socket,
                          None, None, None)

    @mock.patch("murano.common.wsgi.eventlet")
    def test_service_start(self, eventlet):
        self.service = wsgi.Service(None, 1)
        self.service.start()
        self.assertTrue(eventlet.spawn.called)

    def test_service_stop(self):
        self.service = wsgi.Service(None, 1)
        self.service.greenthread = mock.Mock()
        self.service.stop()
        self.assertTrue(self.service.greenthread.kill.called)

    def test_service_properties(self):
        self.service = wsgi.Service(None, 1)
        self.service._socket = mock.Mock()
        self.service._socket.getsockname.return_value = ["host", "port"]
        host = self.service.host
        port = self.service.port
        self.assertEqual("host", host)
        self.assertEqual("port", port)

    @mock.patch("murano.common.wsgi.logging")
    def test_service_reset(self, logging):
        self.service = wsgi.Service(None, 1)
        self.service.reset()
        self.assertTrue(logging.setup.called)

    @mock.patch("murano.common.wsgi.eventlet")
    def test_service_run(self, eventlet):
        self.service = wsgi.Service(None, 1)
        self.service._run(None, None)
        self.assertTrue(eventlet.wsgi.server.called)

    def test_backlog_prop(self):
        service = wsgi.Service(None, None)
        service._backlog = mock.sentinel.backlog
        self.assertEqual(mock.sentinel.backlog, service.backlog)


class TestMiddleware(base.MuranoTestCase):

    def test_middleware_call(self):
        self.middleware = wsgi.Middleware(None)
        mock_request = mock.Mock()
        mock_request.get_response.return_value = "a response"
        self.assertEqual("a response", self.middleware(mock_request))

    def test_call_with_response(self):
        middleware = wsgi.Middleware(None)
        middleware.process_request = mock.Mock(return_value=mock.sentinel.resp)

        resp = middleware(mock.sentinel.req)

        self.assertEqual(mock.sentinel.resp, resp)
        middleware.process_request.assert_called_once_with(mock.sentinel.req)


class TestDebug(base.MuranoTestCase):

    def test_debug_call(self):
        self.debug = wsgi.Debug(None)
        mock_request = mock.Mock()
        mock_request.environ = {"one": 1, "two": 2}

        mock_response = mock.Mock()
        mock_response.headers = {"one": 1, "two": 2}
        mock_request.get_response.return_value = mock_response

        self.assertEqual(mock_response, self.debug(mock_request))

    @mock.patch('sys.stdout')
    def test_print_generator(self, mock_stdout):
        for x in wsgi.Debug.print_generator(['foo', 'bar', 'baz']):
            pass
        mock_stdout.write.assert_has_calls([
            mock.call('**************************************** BODY'),
            mock.call('\n'),
            mock.call('foo'),
            mock.call('bar'),
            mock.call('baz'),
            mock.call(''),
            mock.call('\n')
        ])


class TestRequest(base.MuranoTestCase):

    @mock.patch.object(wsgi.webob.BaseRequest, 'path',
                       **{'rsplit.return_value': ['foo.bar', 'baz']})
    def test_best_match_content_type_with_multi_part_path(self, mock_path):
        request = wsgi.Request({})
        supported_content_types = ['application/baz']
        result = request.best_match_content_type(None, supported_content_types)
        self.assertEqual('application/baz', result)


class TestResource(base.MuranoTestCase):

    def test_resource_call_exceptions(self):
        self.resource = wsgi.Resource(None)
        self.resource.deserialize_request = mock.Mock()
        self.resource.deserialize_request.side_effect = (
            exceptions.UnsupportedContentType
        )
        mock_request = mock.Mock()
        mock_request.headers = {}
        result = self.resource(mock_request)
        self.assertEqual(415, result.status_code)

        self.resource.deserialize_request.side_effect = (
            exceptions.MalformedRequestBody
        )
        result = self.resource(mock_request)
        self.assertEqual(400, result.status_code)

        self.resource.deserialize_request.side_effect = None
        self.resource.deserialize_request.return_value = ["", {"k": "v"}, ""]
        self.resource.execute_action = mock.Mock()
        self.resource.serialize_response = mock.Mock()
        self.resource.serialize_response.side_effect = Exception

        result = self.resource(mock_request)
        self.assertIsNotNone(result)

    def test_get_action_args(self):
        self.resource = wsgi.Resource(None)
        result = self.resource.get_action_args(None)
        self.assertEqual({}, result)

        request_environment = {'wsgiorg.routing_args': ["arg_0", {"k": "v"}]}
        result = self.resource.get_action_args(request_environment)
        self.assertEqual({"k": "v"}, result)

    def test_dispatch_except_attribute_error(self):
        mock_obj = mock.Mock(spec=wsgi.Resource)
        setattr(mock_obj, 'default',
                mock.Mock(return_value=mock.sentinel.ret_value))

        resource = wsgi.Resource(None)
        result = resource.dispatch(mock_obj, 'invalid_action')

        self.assertEqual(mock.sentinel.ret_value, result)
        mock_obj.default.assert_called_once_with()


class TestActionDispatcher(base.MuranoTestCase):

    def test_default(self):
        action_dispatcher = wsgi.ActionDispatcher()
        self.assertRaises(NotImplementedError, action_dispatcher.default, None)


class TestDictSerializer(base.MuranoTestCase):

    def test_default(self):
        dict_serializer = wsgi.DictSerializer()
        self.assertEqual("", dict_serializer.default(None))


class TestXMLDictSerializer(base.MuranoTestCase):

    def test_router_dispatch_not_found(self):
        self.router = wsgi.Router(None)
        req = mock.Mock()
        req.environ = {'wsgiorg.routing_args': [False, False]}
        response = self.router._dispatch(req)
        self.assertIsInstance(response, webob.exc.HTTPNotFound)

    def test_XML_Dict_Serializer_default_string(self):
        self.serializer = wsgi.XMLDictSerializer()
        actual_xml = self.serializer.default({"XML Root": "some data"})
        expected_xml = b'<XML Root>some data</XML Root>\n'
        self.assertEqual(expected_xml, actual_xml)

    def test_XML_Dict_Serializer_node_dict(self):
        self.serializer = wsgi.XMLDictSerializer()
        doc = minidom.Document()
        metadata = mock.Mock()
        metadata.get.return_value = {"node": {"item_name": "name",
                                              "item_key": "key"}}
        nodename = "node"
        data = {"data_key": "data_value"}
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual('<name key="data_key">data_value</name>',
                         result.childNodes[0].toxml())

        mock_get_nodename = mock.Mock()
        mock_get_nodename.get.return_vale = "k"
        metadata.get.return_value = {"k": "v"}
        nodename = "node"
        data = {"data_key": "data_value"}
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual("node", result.nodeName)

        metadata.get.return_value = {}
        nodename = "node"
        data = {"data_key": "data_value"}
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual('<data_key>data_value</data_key>',
                         result.childNodes[0].toxml())

    def test_XML_Dict_Serializer_node_list(self):
        self.serializer = wsgi.XMLDictSerializer()
        doc = minidom.Document()
        metadata = mock.Mock()
        metadata.get.return_value = {"node": {"item_name": "name",
                                              "item_key": "key"}}
        nodename = "node"
        data = ["data_1", "data_2", "data_3"]
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual('<name key="data_1"/>', result.childNodes[0].toxml())

        nodename = "not_node"
        data = ["data_1", "data_2", "data_3"]
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual(3, len(result.childNodes))

        nodename = "nodes"
        data = ["data_1", "data_2", "data_3"]
        result = self.serializer._to_xml_node(doc, metadata, nodename, data)
        self.assertEqual(3, len(result.childNodes))

    def test_XML_Dict_Serializer_create_link_nodes(self):
        self.serializer = wsgi.XMLDictSerializer()
        xml_doc = minidom.Document()
        links = [{"rel": "rel", "href": "href", "type": "type"}]
        link_nodes = self.serializer._create_link_nodes(xml_doc, links)
        result = link_nodes[0].toxml()
        for item in ['<atom:link', 'rel="rel"', 'href="href"', 'type="type"']:
            self.assertIn(item, result)

    def test_add_xmlns(self):
        xml_dict_serializer = wsgi.XMLDictSerializer()
        xml_dict_serializer.xmlns = mock.sentinel.xmlns
        mock_node = mock.Mock()

        xml_dict_serializer._add_xmlns(mock_node, has_atom=False)
        mock_node.setAttribute.assert_called_once_with(
            'xmlns', mock.sentinel.xmlns)

        mock_node.reset_mock()
        xml_dict_serializer._add_xmlns(mock_node, has_atom=True)
        mock_node.setAttribute.assert_has_calls([
            mock.call('xmlns', mock.sentinel.xmlns),
            mock.call('xmlns:atom', "http://www.w3.org/2005/Atom")
        ])


class TestTextDeserializer(base.MuranoTestCase):

    def test_default(self):
        text_deserializer = wsgi.TextDeserializer()
        self.assertEqual({}, text_deserializer.default(None))


class TestJSONDeserializer(base.MuranoTestCase):

    def test_from_json_except_value_error(self):
        json_deserializer = wsgi.JSONDeserializer()
        e = self.assertRaises(exceptions.MalformedRequestBody,
                              json_deserializer._from_json,
                              "throw value error")
        self.assertIn('cannot understand JSON', str(e))


class TestJSONPatchDeserializer(base.MuranoTestCase):

    def test_from_json_patch_except_value_error(self):
        json_patch_deserializer = wsgi.JSONPatchDeserializer()
        e = self.assertRaises(exceptions.MalformedRequestBody,
                              json_patch_deserializer._from_json_patch,
                              "throw value error")
        self.assertIn('cannot understand JSON', str(e))

    def test_validate_allowed_methods_except_http_forbidden(self):
        json_patch_deserializer = wsgi.JSONPatchDeserializer()
        json_patch_deserializer.allowed_operations = mock.Mock(
            spec_set=wsgi.JSONPatchDeserializer.allowed_operations,
            **{'get.return_value': None}
        )

        e = self.assertRaises(
            webob.exc.HTTPForbidden,
            json_patch_deserializer._validate_allowed_methods,
            {'op': 'foo_op', 'path': 'foo_path'}, allow_unknown_path=False)
        self.assertEqual("Attribute 'f/o/o/_/p/a/t/h' is invalid", str(e))

    def test_validate_json_pointer(self):
        json_patch_deserializer = wsgi.JSONPatchDeserializer()
        expected_exception = webob.exc.HTTPBadRequest
        bad_pointers = ['pointer', '/ /\n/', '//', '// ', '/~2']

        for pointer in bad_pointers:
            self.assertRaises(expected_exception,
                              json_patch_deserializer._validate_json_pointer,
                              pointer)

    @mock.patch.object(wsgi, 'jsonschema', autospec=True)
    def test_validate_schema_cannot_validate(self, mock_jsonschema):
        json_patch_deserializer = wsgi.JSONPatchDeserializer()
        json_patch_deserializer.schema = {'type': 'array'}
        test_change = {'path': ['foo_path'], 'value': 'foo_value'}
        # Assert that jsonschema validation was not performed because
        # can_validate = False.
        json_patch_deserializer._validate_schema(test_change)
        self.assertFalse(mock_jsonschema.validate.called)

        mock_jsonschema.validate.reset_mock()
        json_patch_deserializer = wsgi.JSONPatchDeserializer()
        json_patch_deserializer.schema = {'type': 'object'}
        json_patch_deserializer._validate_schema(test_change)
        self.assertFalse(mock_jsonschema.validate.called)


class TestMuranoPackageJSONPatchDeserializer(base.MuranoTestCase):

    def test_validate_path(self):
        patch_deserializer = wsgi.MuranoPackageJSONPatchDeserializer()
        e = self.assertRaises(webob.exc.HTTPBadRequest,
                              patch_deserializer._validate_path,
                              ['foo_path', 'bar_path'])
        self.assertEqual(_('Nested paths are not allowed'), str(e))


class TestXMLDeserializer(base.MuranoTestCase):

    def test_XMLDeserializer_from_xml(self):
        self.deserializer = wsgi.XMLDeserializer()
        datastring = "<XML>HELLO</XML>"
        request = mock.Mock()
        request.body = datastring
        result = self.deserializer._from_xml(request)
        self.assertEqual({'XML': 'HELLO'}, result)

        datastring = "<NOT XML><NOT XML>"
        request.body = datastring
        self.assertRaises(exceptions.MalformedRequestBody,
                          self.deserializer._from_xml, request)
        datastring = "<XML><XML1>HELLO</XML1></XML>"
        request.body = datastring
        result = self.deserializer._from_xml(request)
        self.assertEqual({'XML': {'XML1': 'HELLO'}}, result)

        datastring = "<XML><XML1><XML2>HELLO</XML2></XML1></XML>"
        request.body = datastring
        result = self.deserializer._from_xml(request)
        self.assertEqual({'XML': {'XML1': {'XML2': 'HELLO'}}}, result)

    def test_XMLDeserializer_default(self):
        self.deserializer = wsgi.XMLDeserializer()
        datastring = "<XML>HELLO</XML>"
        request = mock.Mock()
        request.body = datastring

        result = self.deserializer.default(request)
        self.assertEqual({'body': {'XML': 'HELLO'}}, result)

    def test_XMLDeserializer_find_first_child_named(self):
        self.deserializer = wsgi.XMLDeserializer()
        parent = mock.Mock()
        mock_child = mock.Mock()
        mock_child.nodeName = "name"
        parent.childNodes = [mock_child]
        result = self.deserializer.find_first_child_named(parent, "name")
        self.assertIsNotNone(result)

        parent.childNodes = []
        result = self.deserializer.find_first_child_named(parent, "name")
        self.assertIsNone(result)

        mock_child.nodeName = "other_name"
        parent.childNodes = [mock_child]
        result = self.deserializer.find_first_child_named(parent, "name")
        self.assertIsNone(result)

    def test_XMLDeserializer_extract_text(self):
        self.deserializer = wsgi.XMLDeserializer()
        mock_node = mock.Mock()
        mock_child = mock.Mock()
        mock_child.nodeType = mock_child.TEXT_NODE
        mock_node.childNodes = [mock_child]
        result = self.deserializer.extract_text(mock_node)

        mock_child.nodeType = None
        mock_node.childNodes = [mock_child]
        result = self.deserializer.extract_text(mock_node)
        self.assertEqual("", result)

        mock_node.childNodes = []
        result = self.deserializer.extract_text(mock_node)
        self.assertEqual("", result)

    def test_from_xml_node_all_branches(self):
        mock_grandchild_node = mock.MagicMock(**{
            'nodeName': 'qux',
            'nodeType': mock.sentinel.grandchild_node_type,
            'childNodes': [
                mock.Mock(nodeType=3, nodeValue=mock.sentinel.qux_node)
            ]
        })
        mock_child_node = mock.MagicMock(**{
            'nodeName': mock.sentinel.child_node,
            'attributes.keys.return_value': ['foo', 'bar', 'baz'],
            'attributes.__getitem__.side_effect': [
                mock.Mock(nodeValue=mock.sentinel.foo_node),
                mock.Mock(nodeValue=mock.sentinel.bar_node),
                mock.Mock(nodeValue=mock.sentinel.baz_node)
            ],
            'childNodes': [
                mock_grandchild_node
            ],
            'TEXT_NODE': mock.sentinel.child_node_type
        })
        mock_node = mock.Mock(**{
            'nodeName': mock.sentinel.parent_node,
            'childNodes': [
                mock_child_node
            ]
        })
        listnames = [mock.sentinel.parent_node]

        xml_deserializer = wsgi.XMLDeserializer()
        result = xml_deserializer._from_xml_node(mock_node, listnames)
        expected_result = [{
            'foo': mock.sentinel.foo_node,
            'bar': mock.sentinel.bar_node,
            'baz': mock.sentinel.baz_node,
            'qux': mock.sentinel.qux_node
        }]

        self.assertEqual(1, len(result))
        self.assertEqual(len(expected_result[0]), len(result[0]))
        for key, val in expected_result[0].items():
            self.assertEqual(val, result[0][key])

    def test_find_children_named(self):
        child_node = mock.Mock(nodeName=mock.sentinel.node_name)
        parent_node = mock.Mock(childNodes=[child_node])

        result = []
        xml_deserializer = wsgi.XMLDeserializer()
        for n in xml_deserializer.\
                find_children_named(parent_node, mock.sentinel.node_name):
            result.append(n)

        self.assertEqual(1, len(result))
        self.assertEqual(child_node, result[0])


class TestFormDataDeserializer(base.MuranoTestCase):

    @mock.patch.object(wsgi, 'LOG', autospec=True)
    def test_from_json_except_value_error(self, mock_log):
        data_serializer = wsgi.FormDataDeserializer()
        value = data_serializer._from_json('value error')

        self.assertEqual('value error', value)
        mock_log.debug.assert_called_once_with(
            "Trying to deserialize 'value error' to json")
        mock_log.warning.assert_called_once_with(
            'Unable to deserialize to json, using raw text')



#!/usr/bin/env python
#
#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
import sys

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from murano.api.v1 import request_statistics
from murano.common import app_loader
from murano.common import config
from murano.common import policy
from murano.common import server
from murano.common import statservice as stats
from murano.common import wsgi

CONF = cfg.CONF


# If ../murano/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
root = os.path.join(os.path.abspath(__file__), os.pardir, os.pardir, os.pardir)
if os.path.exists(os.path.join(root, 'murano', '__init__.py')):
    sys.path.insert(0, root)


def main():
    try:
        config.parse_args()
        config.set_middleware_defaults()
        request_statistics.init_stats()
        policy.init()

        logging.setup(CONF, 'murano')
        workers = CONF.murano.api_workers
        if not workers:
            workers = processutils.get_worker_count()
        launcher = service.launch(
            CONF, server.ApiService(),
            workers=workers, restart_method='mutate')

        app = app_loader.load_paste_app('murano')
        port, host = (CONF.bind_port, CONF.bind_host)
        launcher.launch_service(wsgi.Service(app, port, host))
        launcher.launch_service(server.NotificationService())
        launcher.launch_service(stats.StatsCollectingService())

        launcher.wait()
    except RuntimeError as e:
        sys.stderr.write("ERROR: %s\n" % e)
        sys.exit(1)


if __name__ == '__main__':
    main()
